# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/user_deployment_settings.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import secret_pb2 as tecton__proto_dot_common_dot_secret__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n0tecton_proto/data/user_deployment_settings.proto\x12\x11tecton_proto.data\x1a tecton_proto/common/secret.proto\"\xfa\x02\n\x16UserDeploymentSettings\x12G\n user_deployment_settings_version\x18\x01 \x01(\x05R\x1duserDeploymentSettingsVersion\x12R\n\x11\x64\x61tabricks_config\x18\x02 \x01(\x0b\x32#.tecton_proto.data.DatabricksConfigH\x00R\x10\x64\x61tabricksConfig\x12T\n\x13user_spark_settings\x18\x03 \x01(\x0b\x32$.tecton_proto.data.UserSparkSettingsR\x11userSparkSettings\x12O\n\x0ftenant_settings\x18\x05 \x01(\x0b\x32&.tecton_proto.data.TenantSettingsProtoR\x0etenantSettingsB\x16\n\x14\x64\x61ta_platform_configJ\x04\x08\x04\x10\x05\"\xdf\x01\n\x10\x44\x61tabricksConfig\x12#\n\rworkspace_url\x18\x01 \x01(\tR\x0cworkspaceUrl\x12\x38\n\tapi_token\x18\x02 \x01(\x0b\x32\x1b.tecton_proto.common.SecretR\x08\x61piToken\x12\x1b\n\tuser_name\x18\x03 \x01(\tR\x08userName\x12*\n\x11user_display_name\x18\x04 \x01(\tR\x0fuserDisplayName\x12#\n\rspark_version\x18\x05 \x01(\tR\x0csparkVersion\"\xd7\x01\n\x11UserSparkSettings\x12\x30\n\x14instance_profile_arn\x18\x01 \x01(\tR\x12instanceProfileArn\x12R\n\nspark_conf\x18\x02 \x03(\x0b\x32\x33.tecton_proto.data.UserSparkSettings.SparkConfEntryR\tsparkConf\x1a<\n\x0eSparkConfEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\x94\x06\n\x13TenantSettingsProto\x12M\n\x14\x63hronosphere_api_key\x18\x01 \x01(\x0b\x32\x1b.tecton_proto.common.SecretR\x12\x63hronosphereApiKey\x12I\n!chronosphere_restrict_label_value\x18\x04 \x01(\tR\x1e\x63hronosphereRestrictLabelValue\x12M\n pseudonymize_amplitude_user_name\x18\x02 \x01(\x08:\x04trueR\x1dpseudonymizeAmplitudeUserName\x12T\n\'enable_user_editing_deployment_settings\x18\x03 \x01(\x08R#enableUserEditingDeploymentSettings\x12-\n\x13okta_admin_group_id\x18\x05 \x01(\tR\x10oktaAdminGroupId\x12+\n\x12okta_user_group_id\x18\x06 \x01(\tR\x0foktaUserGroupId\x12\x39\n\x19\x62\x61se_metadata_service_url\x18\x07 \x01(\tR\x16\x62\x61seMetadataServiceUrl\x12\x37\n\x18\x62\x61se_feature_service_url\x18\x08 \x01(\tR\x15\x62\x61seFeatureServiceUrl\x12:\n\x19spicedb_organization_name\x18\t \x01(\tR\x17spicedbOrganizationName\x12=\n\x1b\x63ustomer_facing_tenant_name\x18\n \x01(\tR\x18\x63ustomerFacingTenantName\x12\x41\n\x0c\x61ws_settings\x18\x0b \x01(\x0b\x32\x1e.tecton_proto.data.AwsSettingsR\x0b\x61wsSettings\x12\x30\n\x14internal_tenant_name\x18\x0c \x01(\tR\x12internalTenantName\"\xa8\x05\n\x0b\x41wsSettings\x12>\n\x0b\x64ynamo_role\x18\x06 \x01(\x0b\x32\x1d.tecton_proto.data.AwsIamRoleR\ndynamoRole\x12_\n\x11\x64ynamo_extra_tags\x18\x08 \x03(\x0b\x32\x33.tecton_proto.data.AwsSettings.DynamoExtraTagsEntryR\x0f\x64ynamoExtraTags\x12\x62\n\x12\x63ompute_extra_tags\x18\t \x03(\x0b\x32\x34.tecton_proto.data.AwsSettings.ComputeExtraTagsEntryR\x10\x63omputeExtraTags\x12\x41\n\x0c\x65mr_settings\x18\x0c \x01(\x0b\x32\x1e.tecton_proto.data.EmrSettingsR\x0b\x65mrSettings\x12Q\n\x12\x64ynamo_table_names\x18\r \x01(\x0b\x32#.tecton_proto.data.DynamoTableNamesR\x10\x64ynamoTableNames\x12]\n\x16object_store_locations\x18\x0e \x01(\x0b\x32\'.tecton_proto.data.ObjectStoreLocationsR\x14objectStoreLocations\x1a\x42\n\x14\x44ynamoExtraTagsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x1a\x43\n\x15\x43omputeExtraTagsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01J\x04\x08\x01\x10\x06J\x04\x08\x07\x10\x08J\x04\x08\n\x10\x0bJ\x04\x08\x0b\x10\x0c\"\x94\x01\n\nAwsIamRole\x12\x19\n\x08role_arn\x18\x01 \x01(\tR\x07roleArn\x12J\n\x11intermediate_role\x18\x02 \x01(\x0b\x32\x1d.tecton_proto.data.AwsIamRoleR\x10intermediateRole\x12\x1f\n\x0b\x65xternal_id\x18\x03 \x01(\tR\nexternalId\"S\n\nS3Location\x12\x12\n\x04path\x18\x01 \x01(\tR\x04path\x12\x31\n\x04role\x18\x02 \x01(\x0b\x32\x1d.tecton_proto.data.AwsIamRoleR\x04role\"\"\n\x0c\x44\x42\x46SLocation\x12\x12\n\x04path\x18\x01 \x01(\tR\x04path\"!\n\x0bGCSLocation\x12\x12\n\x04path\x18\x01 \x01(\tR\x04path\"\xf2\x01\n\x13ObjectStoreLocation\x12@\n\x0bs3_location\x18\x01 \x01(\x0b\x32\x1d.tecton_proto.data.S3LocationH\x00R\ns3Location\x12\x46\n\rdbfs_location\x18\x02 \x01(\x0b\x32\x1f.tecton_proto.data.DBFSLocationH\x00R\x0c\x64\x62\x66sLocation\x12\x43\n\x0cgcs_location\x18\x03 \x01(\x0b\x32\x1e.tecton_proto.data.GCSLocationH\x00R\x0bgcsLocationB\x0c\n\nstore_type\"V\n\x0b\x45mrSettings\x12G\n\x10\x65mr_control_role\x18\x01 \x01(\x0b\x32\x1d.tecton_proto.data.AwsIamRoleR\x0e\x65mrControlRole\"\xa8\x03\n\x10\x44ynamoTableNames\x12*\n\x11\x64\x61ta_table_prefix\x18\x01 \x01(\tR\x0f\x64\x61taTablePrefix\x12*\n\x11status_table_name\x18\x02 \x01(\tR\x0fstatusTableName\x12\x42\n\x1ejob_idempotence_key_table_name\x18\x03 \x01(\tR\x1ajobIdempotenceKeyTableName\x12*\n\x11\x63\x61nary_table_name\x18\x04 \x01(\tR\x0f\x63\x61naryTableName\x12/\n\x14\x64\x65lta_log_table_name\x18\x05 \x01(\tR\x11\x64\x65ltaLogTableName\x12.\n\x13metric_table_prefix\x18\x06 \x01(\tR\x11metricTablePrefix\x12\x34\n\x17\x64\x65lta_log_table_name_v2\x18\x07 \x01(\tR\x13\x64\x65ltaLogTableNameV2\x12\x35\n\x17job_metadata_table_name\x18\x08 \x01(\tR\x14jobMetadataTableName\"\xe0\x07\n\x14ObjectStoreLocations\x12P\n\x0fmaterialization\x18\x01 \x01(\x0b\x32&.tecton_proto.data.ObjectStoreLocationR\x0fmaterialization\x12Y\n\x14streaming_checkpoint\x18\x02 \x01(\x0b\x32&.tecton_proto.data.ObjectStoreLocationR\x13streamingCheckpoint\x12h\n\x1c\x66\x65\x61ture_server_configuration\x18\x03 \x01(\x0b\x32&.tecton_proto.data.ObjectStoreLocationR\x1a\x66\x65\x61tureServerConfiguration\x12I\n\x0c\x66\x65\x61ture_repo\x18\x04 \x01(\x0b\x32&.tecton_proto.data.ObjectStoreLocationR\x0b\x66\x65\x61tureRepo\x12G\n\x0b\x65mr_scripts\x18\x05 \x01(\x0b\x32&.tecton_proto.data.ObjectStoreLocationR\nemrScripts\x12]\n\x16materialization_params\x18\x06 \x01(\x0b\x32&.tecton_proto.data.ObjectStoreLocationR\x15materializationParams\x12S\n\x11intermediate_data\x18\x07 \x01(\x0b\x32&.tecton_proto.data.ObjectStoreLocationR\x10intermediateData\x12\\\n\x16\x66\x65\x61ture_server_logging\x18\x08 \x01(\x0b\x32&.tecton_proto.data.ObjectStoreLocationR\x14\x66\x65\x61tureServerLogging\x12\\\n\x16kafka_credentials_base\x18\t \x01(\x0b\x32&.tecton_proto.data.ObjectStoreLocationR\x14kafkaCredentialsBase\x12\\\n\x16push_api_configuration\x18\n \x01(\x0b\x32&.tecton_proto.data.ObjectStoreLocationR\x14pushApiConfiguration\x12O\n\x0f\x64\x61ta_validation\x18\x0b \x01(\x0b\x32&.tecton_proto.data.ObjectStoreLocationR\x0e\x64\x61taValidationB\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.user_deployment_settings_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _USERSPARKSETTINGS_SPARKCONFENTRY._options = None
  _USERSPARKSETTINGS_SPARKCONFENTRY._serialized_options = b'8\001'
  _AWSSETTINGS_DYNAMOEXTRATAGSENTRY._options = None
  _AWSSETTINGS_DYNAMOEXTRATAGSENTRY._serialized_options = b'8\001'
  _AWSSETTINGS_COMPUTEEXTRATAGSENTRY._options = None
  _AWSSETTINGS_COMPUTEEXTRATAGSENTRY._serialized_options = b'8\001'
  _USERDEPLOYMENTSETTINGS._serialized_start=106
  _USERDEPLOYMENTSETTINGS._serialized_end=484
  _DATABRICKSCONFIG._serialized_start=487
  _DATABRICKSCONFIG._serialized_end=710
  _USERSPARKSETTINGS._serialized_start=713
  _USERSPARKSETTINGS._serialized_end=928
  _USERSPARKSETTINGS_SPARKCONFENTRY._serialized_start=868
  _USERSPARKSETTINGS_SPARKCONFENTRY._serialized_end=928
  _TENANTSETTINGSPROTO._serialized_start=931
  _TENANTSETTINGSPROTO._serialized_end=1719
  _AWSSETTINGS._serialized_start=1722
  _AWSSETTINGS._serialized_end=2402
  _AWSSETTINGS_DYNAMOEXTRATAGSENTRY._serialized_start=2243
  _AWSSETTINGS_DYNAMOEXTRATAGSENTRY._serialized_end=2309
  _AWSSETTINGS_COMPUTEEXTRATAGSENTRY._serialized_start=2311
  _AWSSETTINGS_COMPUTEEXTRATAGSENTRY._serialized_end=2378
  _AWSIAMROLE._serialized_start=2405
  _AWSIAMROLE._serialized_end=2553
  _S3LOCATION._serialized_start=2555
  _S3LOCATION._serialized_end=2638
  _DBFSLOCATION._serialized_start=2640
  _DBFSLOCATION._serialized_end=2674
  _GCSLOCATION._serialized_start=2676
  _GCSLOCATION._serialized_end=2709
  _OBJECTSTORELOCATION._serialized_start=2712
  _OBJECTSTORELOCATION._serialized_end=2954
  _EMRSETTINGS._serialized_start=2956
  _EMRSETTINGS._serialized_end=3042
  _DYNAMOTABLENAMES._serialized_start=3045
  _DYNAMOTABLENAMES._serialized_end=3469
  _OBJECTSTORELOCATIONS._serialized_start=3472
  _OBJECTSTORELOCATIONS._serialized_end=4464
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/consumption/consumption.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n*tecton_proto/consumption/consumption.proto\x12\x18tecton_proto.consumption\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1ctecton_proto/common/id.proto\"\xb9\x03\n\x0f\x43onsumptionInfo\x12\x46\n\x11time_bucket_start\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0ftimeBucketStart\x12%\n\x0eunits_consumed\x18\x02 \x01(\x03R\runitsConsumed\x12\x16\n\x06metric\x18\x03 \x01(\tR\x06metric\x12P\n\x07\x64\x65tails\x18\x04 \x03(\x0b\x32\x36.tecton_proto.consumption.ConsumptionInfo.DetailsEntryR\x07\x64\x65tails\x12\x1b\n\tsource_id\x18\x05 \x01(\tR\x08sourceId\x12?\n\x0f\x66\x65\x61ture_view_id\x18\x06 \x01(\x0b\x32\x17.tecton_proto.common.IdR\rfeatureViewId\x12\x33\n\x16online_read_aws_region\x18\x07 \x01(\tR\x13onlineReadAwsRegion\x1a:\n\x0c\x44\x65tailsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\xff\x01\n\x17\x45nrichedConsumptionInfo\x12T\n\x10\x63onsumption_info\x18\x01 \x01(\x0b\x32).tecton_proto.consumption.ConsumptionInfoR\x0f\x63onsumptionInfo\x12\x34\n\x16\x66\x65\x61ture_view_workspace\x18\x03 \x01(\tR\x14\x66\x65\x61tureViewWorkspace\x12*\n\x11\x66\x65\x61ture_view_name\x18\x04 \x01(\tR\x0f\x66\x65\x61tureViewName\x12&\n\x0f\x66\x65\x61ture_view_id\x18\x05 \x01(\tR\rfeatureViewIdJ\x04\x08\x02\x10\x03\x42\x1a\n\x16\x63om.tecton.consumptionP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.consumption.consumption_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026com.tecton.consumptionP\001'
  _CONSUMPTIONINFO_DETAILSENTRY._options = None
  _CONSUMPTIONINFO_DETAILSENTRY._serialized_options = b'8\001'
  _CONSUMPTIONINFO._serialized_start=136
  _CONSUMPTIONINFO._serialized_end=577
  _CONSUMPTIONINFO_DETAILSENTRY._serialized_start=519
  _CONSUMPTIONINFO_DETAILSENTRY._serialized_end=577
  _ENRICHEDCONSUMPTIONINFO._serialized_start=580
  _ENRICHEDCONSUMPTIONINFO._serialized_end=835
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/basic_info.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.args import diff_options_pb2 as tecton__proto_dot_args_dot_diff__options__pb2
from tecton_proto.args import version_constraints_pb2 as tecton__proto_dot_args_dot_version__constraints__pb2
from tecton_proto.common import analytics_options_pb2 as tecton__proto_dot_common_dot_analytics__options__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"tecton_proto/args/basic_info.proto\x12\x11tecton_proto.args\x1a$tecton_proto/args/diff_options.proto\x1a+tecton_proto/args/version_constraints.proto\x1a+tecton_proto/common/analytics_options.proto\"\x91\x02\n\tBasicInfo\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\'\n\x0b\x64\x65scription\x18\x02 \x01(\tB\x05\x92M\x02\x08\x01R\x0b\x64\x65scription\x12!\n\x05owner\x18\x06 \x01(\tB\x0b\x92M\x02\x08\x01\xf2\xe2\x02\x02\x08\x01R\x05owner\x12\"\n\x06\x66\x61mily\x18\x04 \x01(\tB\n\x92M\x02\x08\x01\x82}\x02\x10\x03R\x06\x66\x61mily\x12\x41\n\x04tags\x18\x05 \x03(\x0b\x32&.tecton_proto.args.BasicInfo.TagsEntryB\x05\x92M\x02\x08\x01R\x04tags\x1a\x37\n\tTagsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01J\x04\x08\x03\x10\x04\x42\x13\n\x0f\x63om.tecton.argsP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.args.basic_info_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.argsP\001'
  _BASICINFO_TAGSENTRY._options = None
  _BASICINFO_TAGSENTRY._serialized_options = b'8\001'
  _BASICINFO.fields_by_name['description']._options = None
  _BASICINFO.fields_by_name['description']._serialized_options = b'\222M\002\010\001'
  _BASICINFO.fields_by_name['owner']._options = None
  _BASICINFO.fields_by_name['owner']._serialized_options = b'\222M\002\010\001\362\342\002\002\010\001'
  _BASICINFO.fields_by_name['family']._options = None
  _BASICINFO.fields_by_name['family']._serialized_options = b'\222M\002\010\001\202}\002\020\003'
  _BASICINFO.fields_by_name['tags']._options = None
  _BASICINFO.fields_by_name['tags']._serialized_options = b'\222M\002\010\001'
  _BASICINFO._serialized_start=186
  _BASICINFO._serialized_end=459
  _BASICINFO_TAGSENTRY._serialized_start=398
  _BASICINFO_TAGSENTRY._serialized_end=453
# @@protoc_insertion_point(module_scope)



import inspect
import re
import site
from pathlib import Path

from tecton import conf
from tecton_core import logger as logger_lib
from tecton_proto.args import repo_metadata_pb2
from tecton_proto.common import id_pb2

# Matches frame strings such as "<string>"
SKIP_FRAME_REGEX = re.compile("\<.*\>")


logger = logger_lib.get_logger(__name__)


def get_current_workspace():
    logger.warn(
        "`tecton.cli.common.get_current_workspace` is deprecated and will be removed in 0.7. Use `tecton.get_current_workspace()` instead."
    )
    return conf.get_or_none("TECTON_WORKSPACE")


def construct_fco_source_info(fco_id: id_pb2.Id) -> repo_metadata_pb2.SourceInfo:
    """Get the SourceInfo (file name and line number) for an FCO.

    How it works:
    - This function assumed it is being called from the constructor of an FCO
    - inspect.stack() returns the call stack (starting with this function)
    - Walk up the stack frames until the first file within a tecton repo (a child of .tecton) is found
    - The first valid tecton repo file is considered the filename of the FCO.
    """
    from tecton_core.repo_file_handler import _maybe_get_repo_root

    excluded_site_pkgs = site.getsitepackages() + [site.getusersitepackages()]

    source_info = repo_metadata_pb2.SourceInfo(fco_id=fco_id)
    repo_root = _maybe_get_repo_root()
    if not repo_root:
        return source_info

    frames = inspect.stack()
    repo_root_path = Path(repo_root)
    for frame in frames:
        if SKIP_FRAME_REGEX.match(frame.frame.f_code.co_filename) is not None:
            continue
        frame_path = Path(frame.frame.f_code.co_filename).resolve()
        if not frame_path.exists():
            continue
        if any(pkg in frame.frame.f_code.co_filename for pkg in excluded_site_pkgs):
            # This filtering is needed in case `tecton` is installed using a virtual
            # environment that's created *inside* the repo root. Without this check,
            # Tecton SDK files would incorrectly be considered a valid repo files
            # and would be listed as the FCO's source filename.
            continue
        if repo_root_path in frame_path.parents:
            rel_filename = frame_path.relative_to(repo_root_path)
            source_info.source_lineno = str(frame.lineno)
            source_info.source_filename = str(rel_filename)
            break
    return source_info


def get_debug(ctx):
    while ctx.parent:
        ctx = ctx.parent
    return ctx.params["debug"]


def get_verbose(ctx):
    while ctx.parent:
        ctx = ctx.parent
    return ctx.params["verbose"]



from typing import Callable
from typing import Optional
from typing import Tuple
from typing import Union

import attrs
import pendulum
from typeguard import typechecked

from tecton_core import conf
from tecton_core import errors
from tecton_core import function_deserialization
from tecton_core import time_utils
from tecton_core.specs import tecton_object_spec
from tecton_core.specs import utils
from tecton_proto.args import data_source_config_pb2
from tecton_proto.args import data_source_pb2 as data_source__args_pb2
from tecton_proto.args import virtual_data_source_pb2 as virtual_data_source__args_pb2
from tecton_proto.common import data_source_type_pb2
from tecton_proto.common import schema_container_pb2
from tecton_proto.common import spark_schema_pb2
from tecton_proto.data import batch_data_source_pb2 as batch_data_source__data_pb2
from tecton_proto.data import stream_data_source_pb2 as stream_data_source__data_pb2
from tecton_proto.data import virtual_data_source_pb2 as virtual_data_source__data_pb2

__all__ = [
    "DataSourceSpec",
    "DataSourceSpecArgsSupplement",
    "BatchSourceSpec",
    "HiveSourceSpec",
    "FileSourceSpec",
    "SparkBatchSourceSpec",
    "RedshiftSourceSpec",
    "SnowflakeSourceSpec",
    "DatetimePartitionColumnSpec",
    "StreamSourceSpec",
    "KinesisSourceSpec",
    "KafkaSourceSpec",
    "SparkStreamSourceSpec",
]


@utils.frozen_strict
class DataSourceSpec(tecton_object_spec.TectonObjectSpec):
    batch_source: Optional["BatchSourceSpec"]
    stream_source: Optional["StreamSourceSpec"]
    schema: Optional[schema_container_pb2.SchemaContainer]
    type: data_source_type_pb2.DataSourceType.ValueType

    # This is used for calling backend validation
    # TODO(jake): use validation args here instead
    data_proto: Optional[virtual_data_source__data_pb2.VirtualDataSource] = attrs.field(
        metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True}
    )

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: virtual_data_source__data_pb2.VirtualDataSource, deserialize_funcs_to_main: bool = False
    ) -> "DataSourceSpec":
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_data_proto(
                proto.virtual_data_source_id, proto.fco_metadata
            ),
            batch_source=create_batch_source_from_data_proto(proto.batch_data_source, deserialize_funcs_to_main),
            stream_source=create_stream_source_from_data_proto(proto.stream_data_source, deserialize_funcs_to_main),
            type=utils.get_field_or_none(proto, "data_source_type"),
            schema=utils.get_field_or_none(proto, "schema"),
            data_proto=proto,
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls,
        proto: virtual_data_source__args_pb2.VirtualDataSourceArgs,
        supplement: "DataSourceSpecArgsSupplement",
    ) -> "DataSourceSpec":
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_args_proto(
                proto.virtual_data_source_id, proto.info
            ),
            batch_source=create_batch_source_from_args_proto(proto, supplement),
            stream_source=create_stream_source_from_args_proto(proto, supplement),
            schema=utils.get_field_or_none(proto, "schema"),
            type=utils.get_field_or_none(proto, "type"),
            data_proto=None,
        )


@attrs.define
class DataSourceSpecArgsSupplement:
    """A data class used for supplementing args protos during DataSourceSpec construction.

    This Python data class can be used to pass non-serializable types (e.g. Python functions) or data that is not
    included in args protos (e.g. schemas) into the DataSourceSpec constructor.
    """

    batch_schema: Optional[spark_schema_pb2.SparkSchema] = attrs.field(default=None)
    stream_schema: Optional[spark_schema_pb2.SparkSchema] = attrs.field(default=None)

    batch_post_processor: Optional[Callable] = attrs.field(default=None)
    stream_post_processor: Optional[Callable] = attrs.field(default=None)

    batch_data_source_function: Optional[Callable] = attrs.field(default=None)
    stream_data_source_function: Optional[Callable] = attrs.field(default=None)


@utils.frozen_strict
class BatchSourceSpec:
    """Base class for batch source specs, e.g. a HiveSourceSpec or SnowflakeSourceSpec."""

    timestamp_field: Optional[str]
    timestamp_format: Optional[str]
    post_processor: Optional[Callable] = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})
    spark_schema: spark_schema_pb2.SparkSchema
    data_delay: Optional[pendulum.Duration]


@typechecked
def create_batch_source_from_data_proto(
    proto: batch_data_source__data_pb2.BatchDataSource, deserialize_funcs_to_main: bool = False
) -> Optional[
    Union["HiveSourceSpec", "SparkBatchSourceSpec", "FileSourceSpec", "RedshiftSourceSpec", "SnowflakeSourceSpec"]
]:
    if proto.HasField("hive_table"):
        return HiveSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    elif proto.HasField("spark_data_source_function"):
        return SparkBatchSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    elif proto.HasField("redshift_db"):
        return RedshiftSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    elif proto.HasField("snowflake"):
        return SnowflakeSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    elif proto.HasField("file"):
        return FileSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    else:
        return None


@typechecked
def create_batch_source_from_args_proto(
    proto: virtual_data_source__args_pb2.VirtualDataSourceArgs, supplement: DataSourceSpecArgsSupplement
) -> Optional[
    Union["HiveSourceSpec", "SparkBatchSourceSpec", "FileSourceSpec", "RedshiftSourceSpec", "SnowflakeSourceSpec"]
]:
    if proto.HasField("hive_ds_config"):
        return HiveSourceSpec.from_args_proto(proto.hive_ds_config, supplement)
    elif proto.HasField("spark_batch_config"):
        return SparkBatchSourceSpec.from_args_proto(proto.spark_batch_config, supplement)
    elif proto.HasField("redshift_ds_config"):
        return RedshiftSourceSpec.from_args_proto(proto.redshift_ds_config, supplement)
    elif proto.HasField("snowflake_ds_config"):
        return SnowflakeSourceSpec.from_args_proto(proto.snowflake_ds_config, supplement)
    elif proto.HasField("file_ds_config"):
        return FileSourceSpec.from_args_proto(proto.file_ds_config, supplement)
    else:
        return None


@utils.frozen_strict
class DatetimePartitionColumnSpec:
    column_name: str
    format_string: str
    minimum_seconds: int

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: batch_data_source__data_pb2.DatetimePartitionColumn
    ) -> "DatetimePartitionColumnSpec":
        return cls(
            column_name=utils.get_field_or_none(proto, "column_name"),
            format_string=utils.get_field_or_none(proto, "format_string"),
            minimum_seconds=utils.get_field_or_none(proto, "minimum_seconds"),
        )

    @classmethod
    @typechecked
    def from_args_proto(cls, proto: data_source__args_pb2.DatetimePartitionColumnArgs) -> "DatetimePartitionColumnSpec":
        # This constructor mirrors backend Kotlin logic: https://github.com/tecton-ai/tecton/blob/6107d76680fa97e6155c6da9ee7d2da47bc3d6a7/java/com/tecton/datamodel/datasource/Utils.kt#L41.
        # In the short-term, divergence between Python and Kotlin will be prevented using integration test coverage. Longer-term, these values should probably be derived client-side only and included in the data source args.
        if proto.datepart == "year":
            format_string = "%Y"
            minimum_seconds = 365 * 24 * 60 * 60
        elif proto.datepart == "month":
            format_string = "%m"
            minimum_seconds = 28 * 24 * 60 * 60
        elif proto.datepart == "day":
            format_string = "%d"
            minimum_seconds = 24 * 60 * 60
        elif proto.datepart == "hour":
            format_string = "%H"
            minimum_seconds = 60 * 60
        elif proto.datepart == "date":
            assert proto.zero_padded or not proto.format_string, "Non-zero padded date strings are not supported."
            format_string = "%Y-%m-%d"
            minimum_seconds = 24 * 60 * 60
        else:
            assert False, f"Unexpected datepart string: {proto.datepart}"

        if proto.format_string:
            format_string = proto.format_string
        elif not proto.zero_padded:
            format_string = format_string.replace("%", "%-")

        return DatetimePartitionColumnSpec(
            column_name=proto.column_name, format_string=format_string, minimum_seconds=minimum_seconds
        )


@utils.frozen_strict
class HiveSourceSpec(BatchSourceSpec):
    database: str
    table: str
    datetime_partition_columns: Tuple[DatetimePartitionColumnSpec, ...]

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: batch_data_source__data_pb2.BatchDataSource, deserialize_funcs_to_main: bool = False
    ) -> "HiveSourceSpec":
        # Needed for canary backwards compatability. Can be deleted on the following release.
        if proto.date_partition_column and not proto.datetime_partition_columns:
            proto.datetime_partition_columns.append(
                batch_data_source__data_pb2.DatetimePartitionColumn(
                    column_name=proto.date_partition_column,
                    format_string="%Y-%m-%d",
                    minimum_seconds=24 * 60 * 60,
                )
            )

        post_processor = None
        if proto.HasField("raw_batch_translator"):
            if deserialize_funcs_to_main:
                post_processor = function_deserialization.from_proto_to_main(proto.raw_batch_translator)
            else:
                post_processor = function_deserialization.from_proto(proto.raw_batch_translator)

        return cls(
            timestamp_field=utils.get_field_or_none(proto.timestamp_column_properties, "column_name"),
            timestamp_format=utils.get_field_or_none(proto.timestamp_column_properties, "format"),
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto, "data_delay"),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            table=utils.get_field_or_none(proto.hive_table, "table"),
            database=utils.get_field_or_none(proto.hive_table, "database"),
            datetime_partition_columns=tuple(
                DatetimePartitionColumnSpec.from_data_proto(column) for column in proto.datetime_partition_columns
            ),
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.HiveDataSourceArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "HiveSourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.common_args.HasField("post_processor"):
            post_processor = function_deserialization.from_proto(proto.common_args.post_processor)
        else:
            post_processor = supplement.batch_post_processor

        return cls(
            timestamp_field=utils.get_field_or_none(proto.common_args, "timestamp_field"),
            timestamp_format=utils.get_field_or_none(proto, "timestamp_format"),
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto.common_args, "data_delay"),
            spark_schema=supplement.batch_schema,
            table=utils.get_field_or_none(proto, "table"),
            database=utils.get_field_or_none(proto, "database"),
            datetime_partition_columns=tuple(
                DatetimePartitionColumnSpec.from_args_proto(column) for column in proto.datetime_partition_columns
            ),
        )


@utils.frozen_strict
class FileSourceSpec(BatchSourceSpec):
    uri: Optional[str]
    file_format: batch_data_source__data_pb2.FileDataSourceFormat.ValueType
    convert_to_glue_format: bool
    schema_uri: Optional[str]
    schema_override: Optional[spark_schema_pb2.SparkSchema]

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: batch_data_source__data_pb2.BatchDataSource, deserialize_funcs_to_main: bool = False
    ) -> "FileSourceSpec":
        post_processor = None
        if proto.HasField("raw_batch_translator"):
            if deserialize_funcs_to_main:
                post_processor = function_deserialization.from_proto_to_main(proto.raw_batch_translator)
            else:
                post_processor = function_deserialization.from_proto(proto.raw_batch_translator)

        return cls(
            timestamp_field=utils.get_field_or_none(proto.timestamp_column_properties, "column_name"),
            timestamp_format=utils.get_field_or_none(proto.timestamp_column_properties, "format"),
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto, "data_delay"),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            uri=utils.get_field_or_none(proto.file, "uri"),
            file_format=utils.get_field_or_none(proto.file, "format"),
            convert_to_glue_format=proto.file.convert_to_glue_format,
            schema_uri=utils.get_field_or_none(proto.file, "schema_uri"),
            schema_override=utils.get_field_or_none(proto.file, "schema_override"),
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.FileDataSourceArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "FileSourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.common_args.HasField("post_processor"):
            post_processor = function_deserialization.from_proto(proto.common_args.post_processor)
        else:
            post_processor = supplement.batch_post_processor

        return cls(
            timestamp_field=utils.get_field_or_none(proto.common_args, "timestamp_field"),
            timestamp_format=utils.get_field_or_none(proto, "timestamp_format"),
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto.common_args, "data_delay"),
            spark_schema=supplement.batch_schema,
            uri=utils.get_field_or_none(proto, "uri"),
            file_format=FileSourceSpec._convert_file_format_string_to_enum(proto.file_format),
            convert_to_glue_format=proto.convert_to_glue_format,
            schema_uri=utils.get_field_or_none(proto, "schema_uri"),
            schema_override=utils.get_field_or_none(proto, "schema_override"),
        )

    @staticmethod
    @typechecked
    def _convert_file_format_string_to_enum(
        file_format: str,
    ) -> batch_data_source__data_pb2.FileDataSourceFormat.ValueType:
        return batch_data_source__data_pb2.FileDataSourceFormat.Value(f"FILE_DATA_SOURCE_FORMAT_{file_format.upper()}")


@utils.frozen_strict
class SparkBatchSourceSpec(BatchSourceSpec):
    function: Callable = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})
    supports_time_filtering: bool

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: batch_data_source__data_pb2.BatchDataSource, deserialize_funcs_to_main: bool = False
    ) -> "SparkBatchSourceSpec":
        function = None
        if proto.spark_data_source_function.HasField("function"):
            if deserialize_funcs_to_main:
                function = function_deserialization.from_proto_to_main(proto.spark_data_source_function.function)
            else:
                function = function_deserialization.from_proto(proto.spark_data_source_function.function)

        return cls(
            timestamp_field=None,
            timestamp_format=None,
            post_processor=None,
            data_delay=utils.get_duration_field_or_none(proto, "data_delay"),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            supports_time_filtering=proto.spark_data_source_function.supports_time_filtering,
            function=function,
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.SparkBatchConfigArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "SparkBatchSourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.HasField("data_source_function"):
            data_source_function = function_deserialization.from_proto(proto.data_source_function)
        else:
            data_source_function = supplement.batch_data_source_function

        return cls(
            timestamp_field=None,
            timestamp_format=None,
            post_processor=None,
            data_delay=utils.get_duration_field_or_none(proto, "data_delay"),
            spark_schema=supplement.batch_schema,
            supports_time_filtering=proto.supports_time_filtering,
            function=data_source_function,
        )


@utils.frozen_strict
class RedshiftSourceSpec(BatchSourceSpec):
    endpoint: str
    table: Optional[str]
    query: Optional[str]
    temp_s3: Optional[str]

    def __attrs_post_init__(self):
        if (self.table and self.query) or (not self.table and not self.query):
            raise errors.REDSHIFT_DS_EITHER_TABLE_OR_QUERY

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: batch_data_source__data_pb2.BatchDataSource, deserialize_funcs_to_main: bool = False
    ) -> "RedshiftSourceSpec":
        post_processor = None
        if proto.HasField("raw_batch_translator"):
            if deserialize_funcs_to_main:
                post_processor = function_deserialization.from_proto_to_main(proto.raw_batch_translator)
            else:
                post_processor = function_deserialization.from_proto(proto.raw_batch_translator)

        return cls(
            timestamp_field=utils.get_field_or_none(proto.timestamp_column_properties, "column_name"),
            timestamp_format=utils.get_field_or_none(proto.timestamp_column_properties, "format"),
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto, "data_delay"),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            endpoint=utils.get_field_or_none(proto.redshift_db, "endpoint"),
            table=utils.get_field_or_none(proto.redshift_db, "table"),
            query=utils.get_field_or_none(proto.redshift_db, "query"),
            temp_s3=utils.get_field_or_none(proto.redshift_db, "temp_s3"),
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.RedshiftDataSourceArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "RedshiftSourceSpec":
        # The temp_s3 directory should be set, and this should have been verified during validation. The one exception
        # is for unit tests (i.e. `test_run()`). In unit tests, validation is skipped and the redshift directory should
        # not need to be set, so fallback to an empty string.
        temp_s3 = conf.get_or_none("SPARK_REDSHIFT_TEMP_DIR") or ""

        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.common_args.HasField("post_processor"):
            post_processor = function_deserialization.from_proto(proto.common_args.post_processor)
        else:
            post_processor = supplement.batch_post_processor

        return cls(
            timestamp_field=utils.get_field_or_none(proto.common_args, "timestamp_field"),
            timestamp_format=None,
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto.common_args, "data_delay"),
            spark_schema=supplement.batch_schema,
            endpoint=utils.get_field_or_none(proto, "endpoint"),
            table=utils.get_field_or_none(proto, "table"),
            query=utils.get_field_or_none(proto, "query"),
            temp_s3=temp_s3,
        )


@utils.frozen_strict
class SnowflakeSourceSpec(BatchSourceSpec):
    database: Optional[str]
    schema: Optional[str]
    warehouse: Optional[str]
    url: Optional[str]
    role: Optional[str]
    table: Optional[str]
    query: Optional[str]

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: batch_data_source__data_pb2.BatchDataSource, deserialize_funcs_to_main: bool = False
    ) -> "SnowflakeSourceSpec":
        post_processor = None
        if proto.HasField("raw_batch_translator"):
            if deserialize_funcs_to_main:
                post_processor = function_deserialization.from_proto_to_main(proto.raw_batch_translator)
            else:
                post_processor = function_deserialization.from_proto(proto.raw_batch_translator)
        return cls(
            timestamp_field=utils.get_field_or_none(proto.timestamp_column_properties, "column_name"),
            timestamp_format=utils.get_field_or_none(proto.timestamp_column_properties, "format"),
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto, "data_delay"),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            url=utils.get_field_or_none(proto.snowflake.snowflakeArgs, "url"),
            database=utils.get_field_or_none(proto.snowflake.snowflakeArgs, "database"),
            schema=utils.get_field_or_none(proto.snowflake.snowflakeArgs, "schema"),
            warehouse=utils.get_field_or_none(proto.snowflake.snowflakeArgs, "warehouse"),
            role=utils.get_field_or_none(proto.snowflake.snowflakeArgs, "role"),
            table=utils.get_field_or_none(proto.snowflake.snowflakeArgs, "table"),
            query=utils.get_field_or_none(proto.snowflake.snowflakeArgs, "query"),
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.SnowflakeDataSourceArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "SnowflakeSourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.common_args.HasField("post_processor"):
            post_processor = function_deserialization.from_proto(proto.common_args.post_processor)
        else:
            post_processor = supplement.batch_post_processor

        return cls(
            timestamp_field=utils.get_field_or_none(proto.common_args, "timestamp_field"),
            timestamp_format=None,
            post_processor=post_processor,
            data_delay=utils.get_duration_field_or_none(proto.common_args, "data_delay"),
            spark_schema=supplement.batch_schema,
            url=utils.get_field_or_none(proto, "url"),
            database=utils.get_field_or_none(proto, "database"),
            schema=utils.get_field_or_none(proto, "schema"),
            warehouse=utils.get_field_or_none(proto, "warehouse"),
            role=utils.get_field_or_none(proto, "role"),
            table=utils.get_field_or_none(proto, "table"),
            query=utils.get_field_or_none(proto, "query"),
        )


@utils.frozen_strict
class StreamOptionSpec:
    key: str
    value: str


@utils.frozen_strict
class StreamSourceSpec:
    """Base class for stream source specs, e.g. a KinesisSourceSpec or KafkaSourceSpec."""

    deduplication_column_names: Tuple[str, ...]
    options: Tuple[StreamOptionSpec, ...]
    watermark_delay_threshold: pendulum.Duration
    spark_schema: spark_schema_pb2.SparkSchema


@typechecked
def create_stream_source_from_data_proto(
    proto: stream_data_source__data_pb2.StreamDataSource, deserialize_funcs_to_main: bool = False
) -> Optional[Union["KinesisSourceSpec", "KafkaSourceSpec", "SparkStreamSourceSpec"]]:
    if proto.HasField("kinesis_data_source"):
        return KinesisSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    elif proto.HasField("kafka_data_source"):
        return KafkaSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    elif proto.HasField("spark_data_source_function"):
        return SparkStreamSourceSpec.from_data_proto(proto, deserialize_funcs_to_main)
    else:
        return None


@typechecked
def create_stream_source_from_args_proto(
    proto: virtual_data_source__args_pb2.VirtualDataSourceArgs, supplement: DataSourceSpecArgsSupplement
) -> Optional[Union["KinesisSourceSpec", "KafkaSourceSpec", "SparkStreamSourceSpec"]]:
    if proto.HasField("kinesis_ds_config"):
        return KinesisSourceSpec.from_args_proto(proto.kinesis_ds_config, supplement)
    elif proto.HasField("kafka_ds_config"):
        return KafkaSourceSpec.from_args_proto(proto.kafka_ds_config, supplement)
    elif proto.HasField("spark_stream_config"):
        return SparkStreamSourceSpec.from_args_proto(proto.spark_stream_config, supplement)
    else:
        return None


@utils.frozen_strict
class KinesisSourceSpec(StreamSourceSpec):
    post_processor: Callable = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})
    stream_name: str
    region: str
    initial_stream_position: data_source_config_pb2.InitialStreamPosition.ValueType

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: stream_data_source__data_pb2.StreamDataSource, deserialize_funcs_to_main: bool = False
    ) -> "KinesisSourceSpec":
        post_processor = None
        if proto.HasField("raw_stream_translator"):
            if deserialize_funcs_to_main:
                post_processor = function_deserialization.from_proto_to_main(proto.raw_stream_translator)
            else:
                post_processor = function_deserialization.from_proto(proto.raw_stream_translator)

        return cls(
            deduplication_column_names=utils.get_tuple_from_repeated_field(proto.deduplication_column_names),
            post_processor=post_processor,
            options=tuple(StreamOptionSpec(key=option.key, value=option.value) for option in proto.options),
            watermark_delay_threshold=time_utils.proto_to_duration(proto.stream_config.watermark_delay_threshold),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            stream_name=utils.get_field_or_none(proto.kinesis_data_source, "stream_name"),
            region=utils.get_field_or_none(proto.kinesis_data_source, "region"),
            initial_stream_position=proto.stream_config.initial_stream_position,
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.KinesisDataSourceArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "KinesisSourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.common_args.HasField("post_processor"):
            post_processor = function_deserialization.from_proto(proto.common_args.post_processor)
        else:
            post_processor = supplement.stream_post_processor

        return cls(
            deduplication_column_names=utils.get_tuple_from_repeated_field(proto.common_args.deduplication_columns),
            post_processor=post_processor,
            options=tuple(StreamOptionSpec(key=option.key, value=option.value) for option in proto.options),
            watermark_delay_threshold=time_utils.proto_to_duration(proto.common_args.watermark_delay_threshold),
            spark_schema=supplement.stream_schema,
            stream_name=utils.get_field_or_none(proto, "stream_name"),
            region=utils.get_field_or_none(proto, "region"),
            initial_stream_position=proto.initial_stream_position,
        )


@utils.frozen_strict
class KafkaSourceSpec(StreamSourceSpec):
    post_processor: Callable = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})
    bootstrap_servers: str
    topics: str
    ssl_keystore_location: Optional[str]
    ssl_keystore_password_secret_id: Optional[str]
    ssl_truststore_location: Optional[str]
    ssl_truststore_password_secret_id: Optional[str]
    security_protocol: Optional[str]

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: stream_data_source__data_pb2.StreamDataSource, deserialize_funcs_to_main: bool = False
    ) -> "KafkaSourceSpec":
        post_processor = None
        if proto.HasField("raw_stream_translator"):
            if deserialize_funcs_to_main:
                post_processor = function_deserialization.from_proto_to_main(proto.raw_stream_translator)
            else:
                post_processor = function_deserialization.from_proto(proto.raw_stream_translator)

        return cls(
            deduplication_column_names=utils.get_tuple_from_repeated_field(proto.deduplication_column_names),
            post_processor=post_processor,
            options=tuple(StreamOptionSpec(key=option.key, value=option.value) for option in proto.options),
            watermark_delay_threshold=time_utils.proto_to_duration(proto.stream_config.watermark_delay_threshold),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            bootstrap_servers=utils.get_field_or_none(proto.kafka_data_source, "bootstrap_servers"),
            topics=utils.get_field_or_none(proto.kafka_data_source, "topics"),
            ssl_keystore_location=utils.get_field_or_none(proto.kafka_data_source, "ssl_keystore_location"),
            ssl_keystore_password_secret_id=utils.get_field_or_none(
                proto.kafka_data_source, "ssl_keystore_password_secret_id"
            ),
            ssl_truststore_location=utils.get_field_or_none(proto.kafka_data_source, "ssl_truststore_location"),
            ssl_truststore_password_secret_id=utils.get_field_or_none(
                proto.kafka_data_source, "ssl_truststore_password_secret_id"
            ),
            security_protocol=utils.get_field_or_none(proto.kafka_data_source, "security_protocol"),
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.KafkaDataSourceArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "KafkaSourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.common_args.HasField("post_processor"):
            post_processor = function_deserialization.from_proto(proto.common_args.post_processor)
        else:
            post_processor = supplement.stream_post_processor

        return cls(
            deduplication_column_names=utils.get_tuple_from_repeated_field(proto.common_args.deduplication_columns),
            post_processor=post_processor,
            options=tuple(StreamOptionSpec(key=option.key, value=option.value) for option in proto.options),
            watermark_delay_threshold=time_utils.proto_to_duration(proto.common_args.watermark_delay_threshold),
            spark_schema=supplement.stream_schema,
            bootstrap_servers=utils.get_field_or_none(proto, "kafka_bootstrap_servers"),
            topics=utils.get_field_or_none(proto, "topics"),
            ssl_keystore_location=utils.get_field_or_none(proto, "ssl_keystore_location"),
            ssl_keystore_password_secret_id=utils.get_field_or_none(proto, "ssl_keystore_password_secret_id"),
            ssl_truststore_location=utils.get_field_or_none(proto, "ssl_truststore_location"),
            ssl_truststore_password_secret_id=utils.get_field_or_none(proto, "ssl_truststore_password_secret_id"),
            security_protocol=utils.get_field_or_none(proto, "security_protocol"),
        )


@utils.frozen_strict
class SparkStreamSourceSpec(StreamSourceSpec):
    function: Callable = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})

    @classmethod
    @typechecked
    def from_data_proto(
        cls, proto: stream_data_source__data_pb2.StreamDataSource, deserialize_funcs_to_main: bool = False
    ) -> "SparkStreamSourceSpec":
        function = None
        if proto.spark_data_source_function.HasField("function"):
            if deserialize_funcs_to_main:
                function = function_deserialization.from_proto_to_main(proto.spark_data_source_function.function)
            else:
                function = function_deserialization.from_proto(proto.spark_data_source_function.function)

        return cls(
            deduplication_column_names=utils.get_tuple_from_repeated_field(proto.deduplication_column_names),
            options=tuple(StreamOptionSpec(key=option.key, value=option.value) for option in proto.options),
            watermark_delay_threshold=time_utils.proto_to_duration(proto.stream_config.watermark_delay_threshold),
            spark_schema=utils.get_field_or_none(proto, "spark_schema"),
            function=function,
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: data_source__args_pb2.SparkStreamConfigArgs, supplement: DataSourceSpecArgsSupplement
    ) -> "SparkStreamSourceSpec":
        # If a function was serialized for this data source (e.g. because it was defined in a repo), then prefer to
        # use the serialized function over directly using the Python function.
        if proto.HasField("data_source_function"):
            data_source_function = function_deserialization.from_proto(proto.data_source_function)
        else:
            data_source_function = supplement.stream_data_source_function

        return cls(
            deduplication_column_names=tuple(),
            options=tuple(),
            watermark_delay_threshold=pendulum.Duration(),
            spark_schema=supplement.stream_schema,
            function=data_source_function,
        )


# Resolve forward type declarations.
attrs.resolve_types(DataSourceSpec, locals(), globals())
attrs.resolve_types(StreamSourceSpec, locals(), globals())



import pendulum

from tecton_core import time_utils
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper as FeatureDefinition


"""
This file contains a utility used for adjusting time based on
feature view configurations.
"""


def get_feature_data_time_limits(
    fd: FeatureDefinition,
    spine_time_limits: pendulum.Period,
) -> pendulum.Period:
    """Returns the feature data time limits based on the spine time limits, taking aggregations into account.

    To get the raw data time limits, you need to use additional information that is on your FilteredSource or other data source inputs.

    Args:
        fd: The feature view for which to compute feature data time limits.
        spine_time_limits: The time limits of the spine; the start time is inclusive and the end time is exclusive.
    """

    if fd.is_feature_table:
        return _feature_table_get_feature_data_time_limits(fd, spine_time_limits)
    elif fd.is_temporal:
        return _temporal_fv_get_feature_data_time_limits(fd, spine_time_limits)
    elif fd.is_temporal_aggregate:
        return _temporal_agg_fv_get_feature_data_time_limits(fd, spine_time_limits)

    # Should never happen!
    raise Exception("Feature definition must be a feature table, temporal FV, or temporal agg FV")


def _feature_table_get_feature_data_time_limits(
    fd: FeatureDefinition,
    spine_time_limits: pendulum.Period,
) -> pendulum.Period:
    """Feature data time limits for a feature table.

    This is `[spine_start_time - ttl, spine_end_time)`
    """

    start_time = spine_time_limits.start
    end_time = spine_time_limits.end

    # Subtract by `serving_ttl` to accommodate for enough feature data for the feature expiration
    if fd.serving_ttl:
        start_time = start_time - fd.serving_ttl

    return end_time - start_time


def _align_left(timestamp: pendulum.DateTime, min_scheduling_interval: pendulum.Duration) -> pendulum.DateTime:
    # min scheduling interval may be zero for continuous
    if min_scheduling_interval.total_seconds() > 0:
        timestamp = time_utils.align_time_downwards(timestamp, min_scheduling_interval)
    return pendulum.instance(timestamp)


def _align_right(timestamp: pendulum.DateTime, min_scheduling_interval: pendulum.Duration) -> pendulum.DateTime:
    """Aligns timestamp to the end of the scheduling interval. If it's already aligned,
    adds scheduling interval's length.

    This function does not take allowed lateness into account.

    :param timestamp: The timestamp in python datetime format
    :return: The timestamp of the greatest aligned time <= timestamp, in seconds.
    """
    aligned_left = _align_left(timestamp, min_scheduling_interval)
    return aligned_left + min_scheduling_interval


def _temporal_fv_get_feature_data_time_limits(
    fd: FeatureDefinition,
    spine_time_limits: pendulum.Period,
) -> pendulum.Period:
    """Feature data time limits for a non aggregate feature view.

    Accounts for:
      * serving ttl
      * batch schedule

    Does NOT account for:
      * data delay
      * batch vs streaming
    """

    start_time = spine_time_limits.start
    end_time = spine_time_limits.end

    # Subtract by `serving_ttl` to accommodate for enough feature data for the feature expiration
    if fd.serving_ttl:
        start_time = start_time - fd.serving_ttl

    # Respect feature_start_time if it's set.
    if fd.feature_start_timestamp:
        start_time = max(start_time, fd.feature_start_timestamp)

    start_time = _align_left(start_time, fd.min_scheduling_interval)

    # Since feature data time interval is open on the right, we need to always strictly align right so that
    # with `batch_schedule = 1h`, time end `04:00:00` will be aligned to `05:00:00`.
    # NOTE: This may be more permissive than 'max_source_data_delay' would allow,
    # but that's okay from a correctness perspective since our as-of join
    # should account for this.
    end_time = _align_right(end_time, fd.min_scheduling_interval)

    # It is possible for the start time to be greater than the end time, for example if the feature start time is later
    # than the spine end time. If this happens, we simply set the start time to equal the end time.
    if start_time > end_time:
        start_time = end_time
    return pendulum.Period(start_time, end_time)


def _temporal_agg_fv_get_feature_data_time_limits(
    fd: FeatureDefinition,
    spine_time_limits: pendulum.Period,
) -> pendulum.Period:
    """Feature data time limits for an aggregate feature view.

    Accounts for:
      * aggregation tile interval
      * data delay

    Does NOT account for:
      * batch vs streaming
    """

    start_time = spine_time_limits.start
    end_time = spine_time_limits.end

    # Respect feature_start_time if it's set.
    if fd.feature_start_timestamp:
        start_time = max(start_time, fd.feature_start_timestamp)

    start_time = _align_left(start_time - fd.max_source_data_delay, fd.min_scheduling_interval)

    # Account for final aggregation needing aggregation window prior to earliest timestamp
    start_time = start_time - time_utils.proto_to_duration(fd.max_aggregation_window)

    # Since feature data time interval is open on the right, we need to always strictly align right so that
    # with `batch_schedule = 1h`, time end `04:00:00` will be aligned to `05:00:00`.
    # NOTE: This may be more permissive than 'max_source_data_delay' would allow,
    # but that's okay from a correctness perspective since our as-of join
    # should account for this.
    end_time = _align_right(end_time, fd.min_scheduling_interval)

    # It is possible for the start time to be greater than the end time, for example if the feature start time is later
    # than the spine end time. If this happens, we simply set the start time to equal the end time.
    if start_time > end_time:
        start_time = end_time
    return pendulum.Period(start_time, end_time)



from datetime import datetime
from typing import Dict
from typing import List
from typing import Optional

import pyspark
import pyspark.sql.functions as F

from tecton_core.aggregation_utils import get_aggregation_column_prefix_from_column_name
from tecton_core.aggregation_utils import get_aggregation_column_prefixes
from tecton_core.aggregation_utils import get_aggregation_function_name
from tecton_core.aggregation_utils import sum_of_squares_column_prefix
from tecton_core.feature_view_utils import resolve_function_name
from tecton_core.query_consts import ANCHOR_TIME
from tecton_proto.common import aggregation_function_pb2 as afpb
from tecton_proto.data.feature_view_pb2 import TrailingTimeWindowAggregation
from tecton_spark.aggregation_plans import AggregationPlan
from tecton_spark.aggregation_plans import get_aggregation_plan
from tecton_spark.time_utils import convert_timestamp_to_epoch
from tecton_spark.time_utils import get_timestamp_in_seconds

WINDOW_COLUMN_NAME = "window"


def _get_feature_partial_aggregations(aggregation_plan: AggregationPlan, feature_name: str):
    column_names = set()
    for column_name, aggregated_column in zip(
        aggregation_plan.materialized_column_names(feature_name),
        aggregation_plan.partial_aggregation_transform(feature_name),
    ):
        if column_name in column_names:
            continue
        column_names.add(column_name)

        yield column_name, aggregated_column.alias(column_name)


def _get_continuous_aggregation_value(column_prefix, feature_input_name):
    if column_prefix == get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_COUNT):
        return F.lit(1).cast("long")
    elif column_prefix in set(
        [
            get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_LAST_DISTINCT_N),
            get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N),
            get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N),
            get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_FIRST_DISTINCT_N),
        ]
    ):
        return F.array(F.col(feature_input_name))
    elif column_prefix == get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_MEAN):
        return F.col(feature_input_name).cast("double")
    elif column_prefix == sum_of_squares_column_prefix:
        return F.pow(feature_input_name, 2)
    else:
        return F.col(feature_input_name)


def _convert_window_to_anchor_time(
    output_df: pyspark.sql.DataFrame,
    is_continuous: bool,
    time_key: str,
    version: int,
    window_start_column_name: Optional[str],
    window_end_column_name: Optional[str],
    convert_window_times_to_epoch: bool,
) -> pyspark.sql.DataFrame:
    def _add_time_column(
        df: pyspark.sql.DataFrame, input_ts_column_name: str, output_column_name: str
    ) -> pyspark.sql.DataFrame:
        col = F.col(input_ts_column_name)
        return df.withColumn(
            output_column_name, convert_timestamp_to_epoch(col, version) if convert_window_times_to_epoch else col
        )

    # For continuous aggregations this will simply be the time key.
    if is_continuous:
        return _add_time_column(output_df, time_key, ANCHOR_TIME)

    # Grouping by Spark Window introduces the "window" struct with "start" and "end" columns.
    # We only need to keep the "start" column as an anchor time.
    anchor_column_name = window_start_column_name if window_start_column_name else ANCHOR_TIME
    output_df = _add_time_column(output_df, f"{WINDOW_COLUMN_NAME}.start", anchor_column_name)

    if window_end_column_name:
        output_df = _add_time_column(output_df, f"{WINDOW_COLUMN_NAME}.end", window_end_column_name)

    return output_df.drop(WINDOW_COLUMN_NAME)


def construct_partial_time_aggregation_df(
    df: pyspark.sql.DataFrame,
    join_keys: List[str],
    time_aggregation: TrailingTimeWindowAggregation,
    version: int,
    window_start_column_name: Optional[str] = None,
    window_end_column_name: Optional[str] = None,
    convert_to_epoch: bool = True,
    aggregation_anchor_time: Optional[datetime] = None,
):
    """Constructs a dataframe that performs partial aggregations on the input dataframe.

    Also removes the default window column that is generated, and replaces it with a start time column, and optionally
    an end time column.

    If the time aggregation is continuous, no aggregations will be performed. Instead, a start time column will be added
    but will simply be the same as the timestamp column in the input dataframe.

    Args:
        df: The input dataframe.
        join_keys: The columns to be grouped by for the partial aggregations.
        time_aggregation: The set of partial aggregations to be performed.
        version: The feature store version.
        window_start_column_name: If specified, the name of the start time column; otherwise it will use the default value "_anchor_time".
        window_end_column_name: If specified, the end time column will be included with this name.
        convert_to_epoch: If True, the window start and end times will be converted to epoch.
        aggregation_anchor_time: If specified, the offset for the aggregation windows.

    Returns:
        A dataframe with the partial aggregations. For example, if `user_id` is the only join key, there are two
        aggregations (min and max), `window_start_column_name` is `tile_start_time`, `window_end_column_name` is
        `tile_end_time`, and `convert_to_epoch` is `False`, then the resulting dataframe might look something like:

        +-------+------------+------------+-------------------+-------------------+
        |user_id|value_min_1d|value_max_1d|    tile_start_time|      tile_end_time|
        +-------+------------+------------+-------------------+-------------------+
        |      1|           7|           9|2022-05-15 00:00:00|2022-05-16 00:00:00|
        |      2|          10|          12|2022-05-16 00:00:00|2022-05-17 00:00:00|
        |      1|          13|          13|2022-05-16 00:00:00|2022-05-17 00:00:00|
        +-------+------------+------------+-------------------+-------------------+
    """
    output_columns = set()
    if not time_aggregation.is_continuous:
        group_by_cols = [F.col(join_key) for join_key in join_keys]
        slide_str = f"{time_aggregation.aggregation_slide_period.seconds} seconds"

        anchor_time_offset_string = None
        if aggregation_anchor_time:
            # Compute the offset from the epoch such that anchor_time aligns to an interval boundary of size
            # aggregation_slide_period. i.e. `epoch + offset + (X * slide_period) = anchor_time`, where X is
            # an integer.
            anchor_time_epoch = get_timestamp_in_seconds(aggregation_anchor_time)
            slide_period_seconds = time_aggregation.aggregation_slide_period.seconds
            anchor_time_offset_seconds = anchor_time_epoch % slide_period_seconds
            anchor_time_offset_string = f"{anchor_time_offset_seconds} seconds"

        window_spec = F.window(time_aggregation.time_key, slide_str, slide_str, anchor_time_offset_string)
        group_by_cols = [window_spec] + group_by_cols
        aggregations = []
        for feature in time_aggregation.features:
            aggregation_plan = get_aggregation_plan(
                feature.function, feature.function_params, time_aggregation.is_continuous, time_aggregation.time_key
            )
            for name, aggregation in _get_feature_partial_aggregations(aggregation_plan, feature.input_feature_name):
                if name in output_columns:
                    continue
                output_columns.add(name)
                aggregations.append(aggregation)
        output_df = df.groupBy(*group_by_cols).agg(*aggregations)
    else:
        columns_to_drop = set()
        for feature in time_aggregation.features:
            column_prefixes = get_aggregation_column_prefixes(feature.function)
            for column_prefix in column_prefixes:
                full_name = f"{column_prefix}_{feature.input_feature_name}"
                if full_name in output_columns:
                    continue
                output_columns.add(full_name)
                df = df.withColumn(
                    full_name, _get_continuous_aggregation_value(column_prefix, feature.input_feature_name)
                )
            columns_to_drop.add(feature.input_feature_name)
        # Drop the original feature columns.
        for column in columns_to_drop:
            df = df.drop(column)
        output_df = df

    output_df = _convert_window_to_anchor_time(
        output_df,
        time_aggregation.is_continuous,
        time_aggregation.time_key,
        version,
        window_start_column_name,
        window_end_column_name,
        convert_to_epoch,
    )
    # TOOD: TEC-12299 drop the `timestamp` column in m13n schema for continuous WAFV
    return output_df


def partial_aggregate_column_renames(
    slide_interval_string: str, trailing_time_window_aggregation: TrailingTimeWindowAggregation
) -> Dict[str, str]:
    """Rename partial aggregate columns to human readable format."""
    # Create a map from intermediate rollup column name to preferred column names.
    renaming_map = dict()
    for feature in trailing_time_window_aggregation.features:
        aggregation_plan = get_aggregation_plan(
            feature.function,
            feature.function_params,
            trailing_time_window_aggregation.is_continuous,
            trailing_time_window_aggregation.time_key,
        )
        for old_name in aggregation_plan.materialized_column_names(feature.input_feature_name):
            aggregation_function_name = resolve_function_name(
                get_aggregation_column_prefix_from_column_name(feature.function, old_name), feature.function_params
            )
            renaming_map[old_name] = f"{feature.input_feature_name}_{aggregation_function_name}_{slide_interval_string}"
    return renaming_map


def rename_partial_aggregate_columns(
    df: pyspark.sql.DataFrame,
    slide_interval_string: str,
    trailing_time_window_aggregation: TrailingTimeWindowAggregation,
) -> pyspark.sql.DataFrame:
    """Rename partial aggregate columns to human readable format."""
    # Create a map from intermediate rollup column name to preferred column names.
    renaming_map = partial_aggregate_column_renames(slide_interval_string, trailing_time_window_aggregation)
    for old_name, new_name in renaming_map.items():
        df = df.withColumnRenamed(old_name, new_name)
    return df



import time
from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import attr
import numpy
import pandas
import pyspark
import pytz

from tecton import snowflake_context
from tecton._internals.data_frame_helper import _get_time_limits_of_dataframe
from tecton._internals.rewrite import rewrite_tree_for_spine
from tecton._internals.sdk_decorators import sdk_public_method
from tecton._internals.utils import get_time_limits_of_pandas_dataframe
from tecton_athena import athena_session
from tecton_athena.data_catalog_helper import register_feature_view_as_athena_table_if_necessary
from tecton_core import conf
from tecton_core.logger import get_logger
from tecton_core.query.node_interface import DataframeWrapper
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.node_interface import recurse_query_tree
from tecton_core.query.nodes import OfflineStoreScanNode
from tecton_core.query.nodes import UserSpecifiedDataNode
from tecton_spark.query import translate

logger = get_logger("Dataframe")

# We have to use Any here because snowflake.snowpark.DataFrame is not a direct dependency of the SDK.
snowpark_dataframe = Any


def set_pandas_timezone_from_spark(pandas_df):
    """Match pandas timezone to that of Spark, s.t. the timestamps are correctly displayed."""
    from tecton.tecton_context import TectonContext

    tz = TectonContext.get_instance()._spark.conf.get("spark.sql.session.timeZone")
    for col in pandas_df.columns:
        if pandas.core.dtypes.common.is_datetime64_dtype(pandas_df[col]):
            pandas_df[col] = pandas_df[col].dt.tz_localize(pytz.timezone(tz))
            pandas_df[col] = pandas_df[col].dt.tz_convert(pytz.timezone("UTC"))
            pandas_df[col] = pandas_df[col].dt.tz_localize(None)
    return pandas_df


@attr.s(auto_attribs=True)
class FeatureVector(object):
    """
    FeatureVector Class.

    A FeatureVector is a representation of a single feature vector. Usage of a FeatureVector typically involves
    extracting the feature vector using ``to_pandas()``, ``to_dict()``, or ``to_numpy()``.

    """

    _names: List[str]
    _values: List[Union[int, str, bytes, float, list]]
    _effective_times: List[Optional[datetime]]
    slo_info: Optional[Dict[str, str]] = None

    @sdk_public_method
    def to_dict(
        self, return_effective_times: bool = False
    ) -> Dict[str, Union[int, str, bytes, float, list, dict, None]]:
        """Turns vector into a Python dict.

        :param: return_effective_times: Whether to return the effective time of the feature.

        :return: A Python dict.
        """
        if return_effective_times:
            return {
                name: {"value": self._values[i], "effective_time": self._effective_times[i]}
                for i, name in enumerate(self._names)
            }

        return dict(zip(self._names, self._values))

    @sdk_public_method
    def to_pandas(self, return_effective_times: bool = False) -> pandas.DataFrame:
        """Turns vector into a Pandas DataFrame.

        :param: return_effective_times: Whether to return the effective time of the feature as part of the DataFrame.

        :return: A Pandas DataFrame.
        """
        if return_effective_times:
            return pandas.DataFrame(
                list(zip(self._names, self._values, self._effective_times)), columns=["name", "value", "effective_time"]
            )

        return pandas.DataFrame([self._values], columns=self._names)

    @sdk_public_method
    def to_numpy(self, return_effective_times: bool = False) -> numpy.array:
        """Turns vector into a numpy array.

        :param: return_effective_times: Whether to return the effective time of the feature as part of the list.

        :return: A numpy array.
        """
        if return_effective_times:
            return numpy.array([self._values, self._effective_times])

        return numpy.array(self._values)

    def _update(self, other: "FeatureVector"):
        self._names.extend(other._names)
        self._values.extend(other._values)
        self._effective_times.extend(other._effective_times)


# NOTE: use repr=False to avoid printing out the underlying dataframes when used in REPL/notebook.
@attr.define(repr=False)
class TectonDataFrame(DataframeWrapper):
    """
    A thin wrapper around Pandas, Spark, and Snowflake dataframes.
    """

    _spark_df: Optional[pyspark.sql.DataFrame] = None
    _pandas_df: Optional[pandas.DataFrame] = None
    # TODO: Change the type to snowflake.snowpark.DataFrame, currently it will
    # fail type checking for our auto generated doc.
    _snowflake_df: Optional[Any] = None
    # should already by optimized
    _querytree: Optional[NodeRef] = attr.field(default=None, repr=lambda x: "TectonQueryTree")
    _temp_table_registered: Optional[set] = None

    @sdk_public_method
    def explain(
        self,
        node_id: bool = True,
        name: bool = True,
        description: bool = True,
        columns: bool = False,
    ):
        """Prints the query tree. Should only be used when this TectonDataFrame is backed by a query tree.

        Args:
            node_id: If True, the unique id associated with each node will be rendered.
            name: If True, the class names of the nodes will be rendered.
            description: If True, the actions of the nodes will be rendered.
            columns: If True, the columns of each node will be rendered as an appendix after tree itself.
        """
        if self._querytree:
            if not name and not description:
                raise RuntimeError("At least one of 'name' or 'description' must be True.")
            if columns and not node_id:
                raise RuntimeError("Can only show columns if 'node_id' is True.")
            print(self._querytree.pretty_str(node_id=node_id, name=name, description=description, columns=columns))
        else:
            print("Explain is only available for TectonDataFrames backed by a query tree.")

    @sdk_public_method
    def to_spark(self) -> pyspark.sql.DataFrame:
        """Returns data as a Spark DataFrame.

        :return: A Spark DataFrame.
        """
        if self._spark_df is not None:
            return self._spark_df
        else:
            from tecton.tecton_context import TectonContext

            tc = TectonContext.get_instance()
            if self._querytree is not None:
                return translate.spark_convert(self._querytree).to_dataframe(tc._spark)
            elif self._pandas_df is not None:
                return tc._spark.createDataFrame(self._pandas_df)
            else:
                raise NotImplementedError

    @sdk_public_method
    def to_pandas(self) -> pandas.DataFrame:
        """Returns data as a Pandas DataFrame.

        :return: A Pandas DataFrame.
        """
        if self._pandas_df is not None:
            return self._pandas_df

        assert self._spark_df is not None or self._snowflake_df is not None or self._querytree is not None

        if self._spark_df is not None:
            return set_pandas_timezone_from_spark(self._spark_df.toPandas())

        if self._snowflake_df is not None:
            return self._snowflake_df.to_pandas(statement_params={"SF_PARTNER": "tecton-ai"})

        if self._querytree is not None:
            if conf.get_or_none("SQL_DIALECT") == "athena":
                session = athena_session.get_session()
                sql = self._to_sql(create_temp_views=True)
                views = self._get_querytree_views()
                for view_name, view_sql in views:
                    session.create_view(view_sql, view_name)
                try:
                    ret = session.read_sql(sql)
                finally:
                    # A tree and its subtree can share the same temp tables. If to_pandas() is called
                    # concurrently, tables can be deleted when they are still being used by the other tree.
                    self._drop_temp_tables()
                    for view_name, _ in views:
                        session.delete_view_if_exists(view_name)
                return ret
            else:
                return set_pandas_timezone_from_spark(self.to_spark().toPandas())

    def _to_sql(self, create_temp_views: bool = False):
        if self._querytree is not None:
            self._temp_table_registered = self._temp_table_registered or set()
            # prevent registering the same spine multiple times.
            if create_temp_views:

                def maybe_register_temp_table(node):
                    if isinstance(node, UserSpecifiedDataNode):
                        tmp_table_name = node.data._temp_table_name
                        if tmp_table_name not in self._temp_table_registered:
                            node.data._register_temp_table()
                        self._temp_table_registered.add(tmp_table_name)

                recurse_query_tree(
                    self._querytree,
                    maybe_register_temp_table,
                )
            # Register Athena table
            if conf.get_or_none("SQL_DIALECT") == "athena":
                session = athena_session.get_session()
                recurse_query_tree(
                    self._querytree,
                    lambda node: register_feature_view_as_athena_table_if_necessary(
                        node.feature_definition_wrapper, session
                    )
                    if isinstance(node, OfflineStoreScanNode)
                    else None,
                )
            return self._querytree.to_sql()
        else:
            raise NotImplementedError

    def _get_querytree_views(self):
        qt_views = []
        recurse_query_tree(self._querytree, lambda node: qt_views.extend(node.get_sql_views()))
        return qt_views

    # TODO(TEC-11097): Deprecate this method after version 0.7.0
    @sdk_public_method
    def to_snowflake(self) -> snowpark_dataframe:
        """
        Deprecated. Please use to_snowpark() instead.
        Returns data as a Snowflake DataFrame.

        :return: A Snowflake DataFrame.
        :meta private:
        """
        logger.warning("to_snowflake() is deprecated - please use to_snowpark() instead.")

        if self._snowflake_df is not None:
            return self._snowflake_df

        assert self._pandas_df is not None

        from tecton.snowflake_context import SnowflakeContext

        if conf.get_bool("ALPHA_SNOWFLAKE_SNOWPARK_ENABLED"):
            return SnowflakeContext.get_instance().get_session().createDataFrame(self._pandas_df)
        else:
            raise Exception("to_snowflake() is only available with Snowpark enabled")

    @sdk_public_method
    def to_snowpark(self) -> snowpark_dataframe:
        """
        Returns data as a Snowpark DataFrame.

        :return: A Snowpark DataFrame.
        """
        if self._snowflake_df is not None:
            return self._snowflake_df

        assert self._pandas_df is not None

        from tecton.snowflake_context import SnowflakeContext

        if conf.get_bool("ALPHA_SNOWFLAKE_SNOWPARK_ENABLED"):
            return SnowflakeContext.get_instance().get_session().createDataFrame(self._pandas_df)
        else:
            raise Exception("to_snowpark() is only available with Snowpark enabled")

    def get_time_range(self, timestamp_key):
        if conf.get_or_none("SQL_DIALECT") == "snowflake":
            raise NotImplementedError
        if conf.get_or_none("SQL_DIALECT") == "athena":
            return get_time_limits_of_pandas_dataframe(self.to_pandas(), timestamp_key)
        else:
            return _get_time_limits_of_dataframe(self.to_spark(), timestamp_key)

    @classmethod
    def _create(
        cls,
        df: Union[pyspark.sql.DataFrame, pandas.DataFrame, NodeRef],
        rewrite: bool = True,
    ):
        """Creates a Tecton DataFrame from a Spark or Pandas DataFrame."""
        if isinstance(df, pandas.DataFrame):
            return cls(spark_df=None, pandas_df=df, snowflake_df=None)
        elif isinstance(df, pyspark.sql.DataFrame):
            return cls(spark_df=df, pandas_df=None, snowflake_df=None)
        elif isinstance(df, NodeRef):
            # TODO: Refactor to not have branching here.
            if rewrite:
                if conf.get_or_none("SQL_DIALECT") == "athena":
                    rewrite_tree_for_spine(df, None)
                else:
                    from tecton.tecton_context import TectonContext

                    spark = TectonContext.get_instance()._spark
                    rewrite_tree_for_spine(df, spark)
            return cls(spark_df=None, pandas_df=None, snowflake_df=None, querytree=df)

        raise TypeError(f"DataFrame must be of type pandas.DataFrame or pyspark.sql.Dataframe, not {type(df)}")

    @classmethod
    # This should be merged into _create once snowpark is installed with pip
    def _create_with_snowflake(cls, df: "snowflake.snowpark.DataFrame"):
        """Creates a Tecton DataFrame from a Snowflake DataFrame."""
        from snowflake.snowpark import DataFrame as SnowflakeDataFrame

        if isinstance(df, SnowflakeDataFrame):
            return cls(spark_df=None, pandas_df=None, snowflake_df=df)

        raise TypeError(f"DataFrame must be of type snowflake.snowpark.Dataframe, not {type(df)}")

    def subtree(self, node_id: int) -> "TectonDataFrame":
        """Creates a TectonDataFrame from a subtree of prior querytree labeled by a node id in .explain()."""
        if not self._querytree:
            raise RuntimeError("Cannot construct a TectonDataFrame from a node id.")

        tree = self._querytree.create_tree()

        # Do not apply rewrites again as they should have already been applied when generating the query tree for this
        # TectonDataFrame.
        return TectonDataFrame._create(NodeRef(tree.get_node(node_id).data), rewrite=False)

    def _timed_to_pandas(self):
        """Convenience method for measuring performance."""
        start = time.time()
        ret = self.to_spark().toPandas()
        end = time.time()
        print(f"took {end-start} seconds")
        return ret

    # The methods below implement the DataframeWrapper interface
    def _register_temp_table(self):
        # TODO(TEC-11647): this will be merged with ALPHA_ATHENA_COMPUTE_ENABLED
        if conf.get_or_none("SQL_DIALECT") == "athena":
            session = athena_session.get_session()
            session.write_pandas(self.to_pandas(), self._temp_table_name)
        elif conf.get_or_none("SQL_DIALECT") == "snowflake":
            session = snowflake_context.SnowflakeContext.get_instance().get_session()
            session.write_pandas(
                self.to_pandas(),
                table_name=self._temp_table_name,
                auto_create_table=True,
                table_type="temporary",
                quote_identifiers=True,
                overwrite=True,
            )
        else:
            self.to_spark().createOrReplaceTempView(self._temp_table_name)

    def _drop_temp_tables(self):
        # TODO(TEC-11647): this will be merged with ALPHA_ATHENA_COMPUTE_ENABLED
        # TODO: implement cleanup using this prior to enabling athena-sql mode in tests
        if conf.get_or_none("SQL_DIALECT") == "athena":
            session = athena_session.get_session()
            if self._temp_table_registered:
                for temp_table in self._temp_table_registered:
                    session.delete_table_if_exists(session.get_database(), temp_table)
                self._temp_table_registered = None
        else:
            from tecton.tecton_context import TectonContext

            spark = TectonContext.get_instance()._spark
            spark.catalog.dropTempView(self._temp_table_name)

    def _select_distinct(self, columns: List[str]) -> DataframeWrapper:
        if self._pandas_df is not None:
            return TectonDataFrame._create(self._pandas_df[columns].drop_duplicates())
        elif self._spark_df is not None:
            return TectonDataFrame._create(self._spark_df.select(columns).distinct())
        else:
            raise NotImplementedError

    @property
    def _dataframe(self) -> Union[pyspark.sql.DataFrame, pandas.DataFrame]:
        if self._spark_df is not None:
            return self._spark_df
        elif self._pandas_df is not None:
            return self._pandas_df
        elif self._querytree:
            if conf.get_or_none("SQL_DIALECT") == "athena":
                return self.to_pandas()
            else:
                return self.to_spark()
        else:
            raise NotImplementedError

    @property
    def columns(self) -> List[str]:
        if self._querytree:
            return self._querytree.columns
        elif self._spark_df is not None:
            return self._spark_df.columns
        elif self._pandas_df is not None:
            return list(self._pandas_df.columns)
        else:
            raise NotImplementedError


# for legacy compat
DataFrame = TectonDataFrame



from base64 import b64encode
from typing import List
from typing import Union

import pandas as pd
from pyspark import sql as pyspark_sql
from pyspark.sql import functions as pyspark_functions
from pyspark.sql import types as pyspark_types

from tecton import tecton_context
from tecton._internals import errors
from tecton._internals import metadata_service
from tecton._internals import spark_api
from tecton._internals import utils
from tecton_core import feature_definition_wrapper
from tecton_core import id_helper
from tecton_proto.common import fco_locator_pb2
from tecton_proto.common import id_pb2
from tecton_proto.metadataservice import metadata_service_pb2
from tecton_proto.online_store import feature_value_pb2
from tecton_spark import ingest_utils


def delete_keys(
    online: bool,
    offline: bool,
    keys: Union[pyspark_sql.DataFrame, pd.DataFrame],
    feature_definition: feature_definition_wrapper.FeatureDefinitionWrapper,
):
    if not offline and not online:
        raise errors.NO_STORE_SELECTED

    materialization_state_transitions = feature_definition.fv_spec.materialization_state_transitions
    if offline and any([transition.offline_enabled for transition in materialization_state_transitions]):
        if not feature_definition.offline_store_config.HasField("delta"):
            raise errors.OFFLINE_STORE_NOT_SUPPORTED

    if online and all([not transition.online_enabled for transition in materialization_state_transitions]):
        print("Online materialization was never enabled. No data to be deleted in online store.")
        online = False

    if offline and all([not transition.offline_enabled for transition in materialization_state_transitions]):
        print("Offline materialization was never enabled. No data to be deleted in offline store.")
        offline = False

    if not (offline or online):
        return None

    spark_df = _get_keys_spark_df(keys, feature_definition.view_schema)
    utils.validate_entity_deletion_keys_dataframe(
        df=spark_df, join_keys=feature_definition.join_keys, view_schema=feature_definition.view_schema
    )

    info_response = _get_delete_entities_info(feature_definition.id)
    s3_path = info_response.df_path
    offline_join_keys_path = s3_path + "/offline"
    online_join_keys_path = s3_path + "/online"
    if online:
        _write_to_online_join_keys_path(online_join_keys_path, keys, feature_definition.join_keys)
    if offline:
        spark_api.write_dataframe_to_path_or_url(
            keys, offline_join_keys_path, info_response.signed_url_for_df_upload_offline, feature_definition.view_schema
        )

    _send_deletion_request(feature_definition, online, offline, online_join_keys_path, offline_join_keys_path)
    return None


def _send_deletion_request(
    feature_definition: feature_definition_wrapper.FeatureDefinitionWrapper,
    online: bool,
    offline: bool,
    online_join_keys_path: str,
    offline_join_keys_path: str,
):
    deletion_request = metadata_service_pb2.DeleteEntitiesRequest(
        fco_locator=fco_locator_pb2.FcoLocator(
            id=id_helper.IdHelper.from_string(feature_definition.id), workspace=feature_definition.workspace
        ),
        online=online,
        offline=offline,
        online_join_keys_path=online_join_keys_path,
        offline_join_keys_path=offline_join_keys_path,
    )
    metadata_service.instance().DeleteEntities(deletion_request)
    print(
        "A deletion job has been created. You can track the status of the job in the Web UI under Materialization section or with deletion_status(). The deletion jobs have a type 'Deletion'."
    )


def _get_delete_entities_info(id_proto: id_pb2.Id) -> metadata_service_pb2.GetDeleteEntitiesInfoResponse:
    info_request = metadata_service_pb2.GetDeleteEntitiesInfoRequest(
        feature_definition_id=id_helper.IdHelper.from_string(id_proto),
    )
    return metadata_service.instance().GetDeleteEntitiesInfo(info_request)


def _get_keys_spark_df(keys: Union[pyspark_sql.DataFrame, pd.DataFrame], view_schema) -> pyspark_sql.DataFrame:
    if isinstance(keys, pd.DataFrame):
        if len(keys) == 0:
            raise errors.EMPTY_ARGUMENT("join_keys")
        if len(keys.columns[keys.columns.duplicated()]):
            raise errors.DUPLICATED_COLS_IN_KEYS(", ".join(list(keys.columns)))
        return ingest_utils.convert_pandas_to_spark_df(keys, view_schema)
    elif isinstance(keys, pyspark_sql.DataFrame):
        return keys
    else:
        raise errors.INVALID_JOIN_KEY_TYPE(type(keys))


def _write_to_online_join_keys_path(
    online_join_keys_path: str, keys: Union[pyspark_sql.DataFrame, pd.DataFrame], join_keys: List[str]
):
    # We actually generate the presigned url but it's not used for online case
    spark = tecton_context.TectonContext.get_instance()._spark
    if isinstance(keys, pd.DataFrame):
        spark_keys_df = spark.createDataFrame(keys)
    else:
        spark_keys_df = keys
    spark_keys_df = spark_keys_df.distinct()
    join_key_df = _serialize_join_keys(spark_keys_df, join_keys)

    # coalesce(1) causes it to write to 1 file, but the jvm code
    # is actually robust to multiple files here
    join_key_df.coalesce(1).write.csv(online_join_keys_path)


def _serialize_join_keys(spark_keys_df: pyspark_sql.DataFrame, join_keys: List[str]):
    def serialize_fn(x):
        ret = feature_value_pb2.FeatureValueList()
        for item in x:
            if isinstance(item, int):
                ret.feature_values.add().int64_value = item
            elif isinstance(item, str):
                ret.feature_values.add().string_value = item
            elif item is None:
                ret.feature_values.add().null_value.CopyFrom(feature_value_pb2.NullValue())
            else:
                raise Exception(f"Unknown type: {type(item)}")
        return b64encode(ret.SerializeToString()).decode()

    serialize = pyspark_functions.udf(serialize_fn, pyspark_types.StringType())
    return spark_keys_df.select(pyspark_functions.struct(*join_keys).alias("join_keys_array")).select(
        serialize("join_keys_array")
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/feature_service.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import framework_version_pb2 as tecton__proto_dot_common_dot_framework__version__pb2
from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2
from tecton_proto.args import basic_info_pb2 as tecton__proto_dot_args_dot_basic__info__pb2
from tecton_proto.args import diff_options_pb2 as tecton__proto_dot_args_dot_diff__options__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\'tecton_proto/args/feature_service.proto\x12\x11tecton_proto.args\x1a+tecton_proto/common/framework_version.proto\x1a\x1ctecton_proto/common/id.proto\x1a\"tecton_proto/args/basic_info.proto\x1a$tecton_proto/args/diff_options.proto\"d\n\x11LoggingConfigArgs\x12\x1f\n\x0bsample_rate\x18\x01 \x01(\x02R\nsampleRate\x12.\n\x13log_effective_times\x18\x02 \x01(\x08R\x11logEffectiveTimes\"V\n\nColumnPair\x12!\n\x0cspine_column\x18\x01 \x01(\tR\x0bspineColumn\x12%\n\x0e\x66\x65\x61ture_column\x18\x02 \x01(\tR\rfeatureColumn\"\xec\x01\n\x1c\x46\x65\x61tureServiceFeaturePackage\x12\x45\n\x12\x66\x65\x61ture_package_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10\x66\x65\x61turePackageId\x12K\n\x12override_join_keys\x18\x02 \x03(\x0b\x32\x1d.tecton_proto.args.ColumnPairR\x10overrideJoinKeys\x12\x1c\n\tnamespace\x18\x03 \x01(\tR\tnamespace\x12\x1a\n\x08\x66\x65\x61tures\x18\x04 \x03(\tR\x08\x66\x65\x61tures\"\xe5\x03\n\x12\x46\x65\x61tureServiceArgs\x12\x45\n\x12\x66\x65\x61ture_service_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10\x66\x65\x61tureServiceId\x12\x37\n\x04info\x18\x02 \x01(\x0b\x32\x1c.tecton_proto.args.BasicInfoB\x05\x92M\x02\x10\x01R\x04info\x12\x46\n\x07version\x18\x06 \x01(\x0e\x32%.tecton_proto.common.FrameworkVersionB\x05\x92M\x02\x08\x05R\x07version\x12.\n\x0fprevent_destroy\x18\x07 \x01(\x08\x42\x05\x92M\x02\x08\x01R\x0epreventDestroy\x12Z\n\x10\x66\x65\x61ture_packages\x18\x03 \x03(\x0b\x32/.tecton_proto.args.FeatureServiceFeaturePackageR\x0f\x66\x65\x61turePackages\x12;\n\x16online_serving_enabled\x18\x04 \x01(\x08\x42\x05\x92M\x02\x08\x01R\x14onlineServingEnabled\x12>\n\x07logging\x18\x05 \x01(\x0b\x32$.tecton_proto.args.LoggingConfigArgsR\x07loggingB\x13\n\x0f\x63om.tecton.argsP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.args.feature_service_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.argsP\001'
  _FEATURESERVICEARGS.fields_by_name['info']._options = None
  _FEATURESERVICEARGS.fields_by_name['info']._serialized_options = b'\222M\002\020\001'
  _FEATURESERVICEARGS.fields_by_name['version']._options = None
  _FEATURESERVICEARGS.fields_by_name['version']._serialized_options = b'\222M\002\010\005'
  _FEATURESERVICEARGS.fields_by_name['prevent_destroy']._options = None
  _FEATURESERVICEARGS.fields_by_name['prevent_destroy']._serialized_options = b'\222M\002\010\001'
  _FEATURESERVICEARGS.fields_by_name['online_serving_enabled']._options = None
  _FEATURESERVICEARGS.fields_by_name['online_serving_enabled']._serialized_options = b'\222M\002\010\001'
  _LOGGINGCONFIGARGS._serialized_start=211
  _LOGGINGCONFIGARGS._serialized_end=311
  _COLUMNPAIR._serialized_start=313
  _COLUMNPAIR._serialized_end=399
  _FEATURESERVICEFEATUREPACKAGE._serialized_start=402
  _FEATURESERVICEFEATUREPACKAGE._serialized_end=638
  _FEATURESERVICEARGS._serialized_start=641
  _FEATURESERVICEARGS._serialized_end=1126
# @@protoc_insertion_point(module_scope)



import io
import os
import sys
import tarfile
import time
from pathlib import Path
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import requests
from google.protobuf.json_format import MessageToJson
from yaspin.spinners import Spinners

import tecton
from .cli_utils import confirm_or_exit
from .error_utils import format_server_errors
from tecton._internals import metadata_service
from tecton._internals.analytics import StateUpdateEventMetrics
from tecton._internals.analytics import StateUpdateResponse
from tecton._internals.utils import is_live_workspace
from tecton.cli import printer
from tecton.cli.engine_renderer import ThickPlanRenderingClient
from tecton.cli.engine_renderer import ThinPlanRenderingClient
from tecton.framework import base_tecton_object
from tecton_core.errors import TectonAPIValidationError
from tecton_core.errors import TectonInternalError
from tecton_core.feature_definition_wrapper import FrameworkVersion
from tecton_core.id_helper import IdHelper
from tecton_proto.args.entity_pb2 import EntityArgs
from tecton_proto.args.fco_args_pb2 import FcoArgs
from tecton_proto.args.feature_service_pb2 import FeatureServiceArgs
from tecton_proto.args.feature_view_pb2 import FeatureViewArgs
from tecton_proto.args.repo_metadata_pb2 import FeatureRepoSourceInfo
from tecton_proto.args.transformation_pb2 import TransformationArgs
from tecton_proto.args.virtual_data_source_pb2 import VirtualDataSourceArgs
from tecton_proto.data.state_update_pb2 import ValidationMessage
from tecton_proto.metadataservice.metadata_service_pb2 import ApplyStateUpdateRequest
from tecton_proto.metadataservice.metadata_service_pb2 import NewStateUpdateRequest
from tecton_proto.metadataservice.metadata_service_pb2 import NewStateUpdateRequestV2
from tecton_proto.metadataservice.metadata_service_pb2 import QueryStateUpdateRequest
from tecton_proto.metadataservice.metadata_service_pb2 import QueryStateUpdateRequestV2
from tecton_proto.metadataservice.metadata_service_pb2 import QueryStateUpdateResponse
from tecton_proto.metadataservice.metadata_service_pb2 import QueryStateUpdateResponseV2


def get_declared_fco_args(objects) -> Tuple[List[FcoArgs], FeatureRepoSourceInfo]:
    all_args = []
    repo_source_info = FeatureRepoSourceInfo()

    for fco_obj in objects:
        # TODO(TEC-12828): Clean this logic up and improve type annotations.
        if isinstance(fco_obj, base_tecton_object.BaseTectonObject):
            all_args.append(fco_obj._build_args())
            repo_source_info.source_info.append(fco_obj._source_info)
        else:
            fco_args = FcoArgs()
            args = fco_obj._args
            source_info = fco_obj._source_info
            if isinstance(args, VirtualDataSourceArgs):
                source_info.fco_id.CopyFrom(args.virtual_data_source_id)
                fco_args.virtual_data_source.CopyFrom(args)
            elif isinstance(args, EntityArgs):
                source_info.fco_id.CopyFrom(args.entity_id)
                fco_args.entity.CopyFrom(args)
            elif isinstance(args, TransformationArgs):
                source_info.fco_id.CopyFrom(args.transformation_id)
                fco_args.transformation.CopyFrom(args)
            elif isinstance(args, FeatureViewArgs):
                source_info.fco_id.CopyFrom(args.feature_view_id)
                fco_args.feature_view.CopyFrom(args)
            elif isinstance(args, FeatureServiceArgs):
                source_info.fco_id.CopyFrom(args.feature_service_id)
                fco_args.feature_service.CopyFrom(args)
            else:
                raise RuntimeError(f"Unknown object {fco_obj}")

            all_args.append(fco_args)

            source_info.CopyFrom(source_info)
            repo_source_info.source_info.append(source_info)

    return all_args, repo_source_info


def dump_local_state(objects):
    with printer.safe_yaspin(Spinners.earth, text="Collecting local feature declarations") as sp:
        fco_args, repo_source_info = get_declared_fco_args(objects)
        sp.ok(printer.safe_string("✅"))

    request_plan = NewStateUpdateRequest()
    request_plan.request.fco_args.extend(fco_args)
    request_plan.request.repo_source_info.CopyFrom(repo_source_info)
    printer.safe_print(MessageToJson(request_plan, including_default_value_fields=True))


# upload tar.gz of python files to url via PUT request
def upload_files(repo_files: List[Path], repo_root, url: str):
    tar_bytes = io.BytesIO()
    with tarfile.open(fileobj=tar_bytes, mode="w|gz") as targz:
        for f in repo_files:
            targz.add(f, arcname=os.path.relpath(f, repo_root))
    for _ in range(3):
        try:
            r = requests.put(url, data=tar_bytes.getbuffer())
            if r.status_code != 200:
                # We will get 403 (forbidden) when the signed url expires.
                if r.status_code == 403:
                    printer.safe_print(
                        f"\nUploading feature repo failed due to expired session. Please retry the command."
                    )
                else:
                    printer.safe_print(f"\nUploading feature repo failed with reason: {r.reason}")
                sys.exit(1)
            return
        except requests.RequestException as e:
            last_error = e
    else:
        raise SystemExit(last_error)


def _construct_new_plan_request_v1(
    workspace_name: str,
    upgrade_all: bool,
    fco_args: List[FcoArgs],
    repo_source_info: FeatureRepoSourceInfo,
    suppress_recreates: bool,
    debug: bool,
) -> NewStateUpdateRequest:
    request_plan = NewStateUpdateRequest()
    request_plan.request.workspace = workspace_name

    request_plan.request.upgrade_all = upgrade_all
    request_plan.request.sdk_version = tecton.version.get_semantic_version() or ""

    request_plan.request.fco_args.extend(fco_args)
    request_plan.request.repo_source_info.CopyFrom(repo_source_info)
    request_plan.request.suppress_recreates = suppress_recreates

    if debug:
        request_plan.blocking_dry_run_mode = True
    else:
        request_plan.enable_eager_response = True

    return request_plan


def _construct_new_plan_request_v2_from_v1(
    request_v1: NewStateUpdateRequest,
    no_color: bool,
    json_out: bool,
) -> NewStateUpdateRequestV2:
    request_plan = NewStateUpdateRequestV2()
    request_plan.request.CopyFrom(request_v1.request)
    request_plan.blocking_dry_run_mode = request_v1.blocking_dry_run_mode
    request_plan.enable_eager_response = request_v1.enable_eager_response
    request_plan.no_color = no_color
    request_plan.json_output = json_out

    return request_plan


def _construct_query_plan_request_v2_from_v1(
    request_v1: QueryStateUpdateRequest,
    no_color: bool,
    json_out: bool,
) -> QueryStateUpdateRequestV2:
    query_request = QueryStateUpdateRequestV2()
    query_request.state_id.CopyFrom(request_v1.state_id)
    query_request.workspace = request_v1.workspace
    query_request.no_color = no_color
    query_request.json_output = json_out

    return query_request


def _execute_new_state_update_v1_or_v2(
    request_v1: NewStateUpdateRequest,
    enable_server_side_rendering: bool,
    no_color: bool,
    json_out: bool,
):
    if enable_server_side_rendering:
        request_v2 = _construct_new_plan_request_v2_from_v1(request_v1, no_color, json_out)
        return metadata_service.instance().NewStateUpdateV2(request_v2)
    else:
        return metadata_service.instance().NewStateUpdate(request_v1)


def _execute_query_state_update_v1_or_v2(
    request_v1: QueryStateUpdateRequest,
    enable_server_side_rendering: bool,
    no_color: bool,
    json_out: bool,
):
    if enable_server_side_rendering:
        request_v2 = _construct_query_plan_request_v2_from_v1(request_v1, no_color, json_out)
        return metadata_service.instance().QueryStateUpdateV2(request_v2)
    else:
        return metadata_service.instance().QueryStateUpdate(request_v1)


def _get_validation_errors(
    query_response: Union[QueryStateUpdateResponse, QueryStateUpdateResponseV2], enable_server_side_rendering: bool
) -> List[ValidationMessage]:
    if enable_server_side_rendering:
        if query_response.validation_errors.errors:
            return query_response.validation_errors.errors
    else:
        if query_response.validation_result.errors:
            return query_response.validation_result.errors
    return []


def update_tecton_state(
    objects,
    repo_files: List[Path],
    repo_root: str,
    apply,
    debug,
    interactive,
    upgrade_all: bool,
    workspace_name: str,
    suppress_warnings: bool = False,
    suppress_recreates: bool = False,
    json_out_path: Optional[Path] = None,
    timeout_seconds=90 * 60,
    plan_id: Optional[str] = None,
    no_color: bool = False,  # used for plan rendering on server-side
    enable_server_side_rendering: bool = True,
) -> StateUpdateResponse:
    # In debug mode we compute the plan synchronously, do not save it in the database, and do not allow to apply it.
    # Primary goal is allowing local development/debugging plans against remote clusters in read-only mode.
    assert not (debug and apply), "Cannot apply in debug mode"
    json_out = json_out_path != None

    if apply and plan_id:
        # Applying an existing plan, so skip preparing args.
        state_id = IdHelper.from_string(plan_id)
        request_query = QueryStateUpdateRequest()
        request_query.state_id.CopyFrom(state_id)
        request_query.workspace = workspace_name

        try:
            response_query = _execute_query_state_update_v1_or_v2(
                request_query,
                enable_server_side_rendering,
                no_color,
                json_out,
            )
        except (
            TectonInternalError,
            TectonAPIValidationError,
        ) as e:
            printer.safe_print(e)
            return StateUpdateResponse.from_error_message(str(e), suppress_recreates)

        if response_query.error:
            printer.safe_print(response_query.error)
            return StateUpdateResponse.from_error_message(response_query.error, suppress_recreates)
        if len(_get_validation_errors(response_query, enable_server_side_rendering)) > 0:
            # Cannot pretty-print validation result using format_server_errors(), because collected local objects
            # might have changed since this plan was generated, so can't accurately match with this plan's FCOs.
            message = f"Cannot apply plan because it had errors."
            printer.safe_print(message)
            return StateUpdateResponse.from_error_message(message, suppress_recreates)

    else:
        with printer.safe_yaspin(Spinners.earth, text="Collecting local feature declarations") as sp:
            fco_args, repo_source_info = get_declared_fco_args(objects)
            sp.ok(printer.safe_string("✅"))

        new_plan_request = _construct_new_plan_request_v1(
            workspace_name, upgrade_all, fco_args, repo_source_info, suppress_recreates, debug
        )

        server_side_msg_prefix = "Performing server-side feature validation"
        with printer.safe_yaspin(Spinners.earth, text=f"{server_side_msg_prefix}: Initializing.") as sp:
            try:
                new_plan_response = _execute_new_state_update_v1_or_v2(
                    new_plan_request,
                    enable_server_side_rendering,
                    no_color,
                    json_out,
                )

                if new_plan_response.HasField("signed_url_for_repo_upload"):
                    upload_files(repo_files, repo_root, new_plan_response.signed_url_for_repo_upload)
                if new_plan_response.HasField("eager_response"):
                    response_query = new_plan_response.eager_response
                else:
                    seconds_slept = 0
                    request_query = QueryStateUpdateRequest()
                    request_query.workspace = workspace_name
                    request_query.state_id.CopyFrom(new_plan_response.state_id)
                    while True:
                        response_query = _execute_query_state_update_v1_or_v2(
                            request_query,
                            enable_server_side_rendering,
                            no_color,
                            json_out,
                        )
                        if response_query.latest_status_message:
                            sp.text = f"{server_side_msg_prefix}: {response_query.latest_status_message}"
                        if response_query.ready:
                            break
                        seconds_to_sleep = 5
                        time.sleep(seconds_to_sleep)
                        seconds_slept += seconds_to_sleep
                        if seconds_slept > timeout_seconds:
                            sp.fail(printer.safe_string("⛔"))
                            printer.safe_print("Validation timed out.")
                            return StateUpdateResponse.from_error_message("Validation timed out.", suppress_recreates)

                if response_query.error:
                    sp.fail(printer.safe_string("⛔"))
                    printer.safe_print(response_query.error)
                    return StateUpdateResponse.from_error_message(response_query.error, suppress_recreates)
                validation_errors = _get_validation_errors(response_query, enable_server_side_rendering)
                if len(validation_errors) > 0:
                    sp.fail(printer.safe_string("⛔"))
                    format_server_errors(validation_errors, objects, repo_root)
                    return StateUpdateResponse.from_error_message(str(validation_errors), suppress_recreates)
                sp.ok(printer.safe_string("✅"))
            except (
                TectonInternalError,
                TectonAPIValidationError,
            ) as e:
                sp.fail(printer.safe_string("⛔"))
                printer.safe_print(e)
                return StateUpdateResponse.from_error_message(str(e), suppress_recreates)

        state_id = new_plan_response.state_id

    if enable_server_side_rendering:
        plan_rendering_client = ThinPlanRenderingClient(response_query)
    else:
        plan_rendering_client = ThickPlanRenderingClient(
            response_query.diff_items,
            response_query.validation_result.warnings,
            state_id,
            suppress_warnings,
            workspace_name,
            is_live_workspace(workspace_name),
            debug,
            response_query.recreates_suppressed,
        )

    if not plan_rendering_client.has_diffs():
        plan_rendering_client.print_empty_plan()
    else:
        plan_rendering_client.print_plan()

        if apply:
            plan_rendering_client.print_apply_warnings()
            if interactive:
                confirm_or_exit(f'Are you sure you want to apply this plan to: "{workspace_name}"?')

            request_apply = ApplyStateUpdateRequest()
            request_apply.state_id.CopyFrom(state_id)
            metadata_service.instance().ApplyStateUpdate(request_apply)

            printer.safe_print("🎉 all done!")

    if json_out_path:
        repo_diff_summary = plan_rendering_client.get_json_plan_output()
        json_out_path.parent.mkdir(parents=True, exist_ok=True)
        json_out_path.write_text(repo_diff_summary)

    num_v3_fcos = sum(
        obj._args.version in (FrameworkVersion.UNSPECIFIED.value, FrameworkVersion.FWV3.value) for obj in objects
    )
    num_v5_fcos = sum(obj._args.version is FrameworkVersion.FWV5.value for obj in objects)

    v1_response = None
    v2_response = None
    if enable_server_side_rendering:
        v2_response = response_query
    else:
        v1_response = response_query

    return StateUpdateResponse(
        state_update_event_metrics=StateUpdateEventMetrics(
            num_total_fcos=len(objects),
            num_fcos_changed=plan_rendering_client.num_fcos_changed,
            num_v3_fcos=num_v3_fcos,
            num_v5_fcos=num_v5_fcos,
            suppress_recreates=suppress_recreates,
            json_out=(json_out_path is not None),
            num_warnings=plan_rendering_client.num_warnings,
        ),
        v1_response=v1_response,
        v2_response=v2_response,
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/materialization/spark_cluster.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.spark_api import jobs_pb2 as tecton__proto_dot_spark__api_dot_jobs__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n0tecton_proto/materialization/spark_cluster.proto\x12\x1ctecton_proto.materialization\x1a!tecton_proto/spark_api/jobs.proto\"\xad\x02\n\x17SparkClusterEnvironment\x12I\n!spark_cluster_environment_version\x18\x01 \x01(\x05R\x1esparkClusterEnvironmentVersion\x12\x65\n\x15job_request_templates\x18\n \x01(\x0b\x32\x31.tecton_proto.materialization.JobRequestTemplatesR\x13jobRequestTemplates\x12T\n\'merged_user_deployment_settings_version\x18\x0c \x01(\x05R#mergedUserDeploymentSettingsVersionJ\x04\x08\x02\x10\nJ\x04\x08\x0b\x10\x0c\"\xc1\x01\n\x13JobRequestTemplates\x12X\n\x13\x64\x61tabricks_template\x18\x02 \x01(\x0b\x32\'.tecton_proto.spark_api.StartJobRequestR\x12\x64\x61tabricksTemplate\x12J\n\x0c\x65mr_template\x18\x03 \x01(\x0b\x32\'.tecton_proto.spark_api.StartJobRequestR\x0b\x65mrTemplateJ\x04\x08\x01\x10\x02*b\n\x08TaskType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\t\n\x05\x42\x41TCH\x10\x01\x12\r\n\tSTREAMING\x10\x02\x12\n\n\x06INGEST\x10\x03\x12\x0c\n\x08\x44\x45LETION\x10\x04\x12\x15\n\x11\x44\x45LTA_MAINTENANCE\x10\x05*x\n\x14\x45xecutionEnvironment\x12\x13\n\x0f\x45NV_UNSPECIFIED\x10\x00\x12\x1b\n\x17\x45NV_DATABRICKS_NOTEBOOK\x10\x01\x12\x0b\n\x07\x45NV_EMR\x10\x02\x12\x11\n\rENV_SNOWFLAKE\x10\x03\x12\x0e\n\nENV_ATHENA\x10\x04\x42\x1e\n\x1a\x63om.tecton.materializationP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.materialization.spark_cluster_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\032com.tecton.materializationP\001'
  _TASKTYPE._serialized_start=617
  _TASKTYPE._serialized_end=715
  _EXECUTIONENVIRONMENT._serialized_start=717
  _EXECUTIONENVIRONMENT._serialized_end=837
  _SPARKCLUSTERENVIRONMENT._serialized_start=118
  _SPARKCLUSTERENVIRONMENT._serialized_end=419
  _JOBREQUESTTEMPLATES._serialized_start=422
  _JOBREQUESTTEMPLATES._serialized_end=615
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/transformation.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.args import user_defined_function_pb2 as tecton__proto_dot_args_dot_user__defined__function__pb2
from tecton_proto.data import fco_metadata_pb2 as tecton__proto_dot_data_dot_fco__metadata__pb2
from tecton_proto.args import transformation_pb2 as tecton__proto_dot_args_dot_transformation__pb2
from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2
from tecton_proto.validation import validator_pb2 as tecton__proto_dot_validation_dot_validator__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&tecton_proto/data/transformation.proto\x12\x11tecton_proto.data\x1a-tecton_proto/args/user_defined_function.proto\x1a$tecton_proto/data/fco_metadata.proto\x1a&tecton_proto/args/transformation.proto\x1a\x1ctecton_proto/common/id.proto\x1a\'tecton_proto/validation/validator.proto\"\x9e\x03\n\x0eTransformation\x12\x44\n\x11transformation_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10transformationId\x12\x41\n\x0c\x66\x63o_metadata\x18\x02 \x01(\x0b\x32\x1e.tecton_proto.data.FcoMetadataR\x0b\x66\x63oMetadata\x12V\n\x13transformation_mode\x18\x03 \x01(\x0e\x32%.tecton_proto.args.TransformationModeR\x12transformationMode\x12K\n\ruser_function\x18\x04 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x0cuserFunction\x12^\n\x0fvalidation_args\x18\x05 \x01(\x0b\x32\x35.tecton_proto.validation.TransformationValidationArgsR\x0evalidationArgsB\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.transformation_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _TRANSFORMATION._serialized_start=258
  _TRANSFORMATION._serialized_end=672
# @@protoc_insertion_point(module_scope)



import functools
import itertools
import os
import random
import time
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from datetime import datetime
from datetime import timedelta
from typing import List
from typing import Optional

import pendulum
from py4j.protocol import Py4JJavaError
from pyspark.sql import Column
from pyspark.sql import DataFrame
from pyspark.sql import functions
from pyspark.sql import SparkSession
from pyspark.sql.types import IntegerType
from pyspark.sql.types import LongType
from pyspark.sql.types import StructType
from pyspark.sql.types import TimestampType

from tecton_core import time_utils as core_time_utils
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper as FeatureDefinition
from tecton_core.logger import get_logger
from tecton_core.offline_store import _check_supported_offline_store_version
from tecton_core.offline_store import _timestamp_formats
from tecton_core.offline_store import partition_col_for_parquet
from tecton_core.offline_store import partition_size_for_delta
from tecton_core.offline_store import partition_size_for_parquet
from tecton_core.offline_store import TIME_PARTITION
from tecton_core.offline_store import window_size_seconds
from tecton_core.query_consts import ANCHOR_TIME
from tecton_spark import time_utils as spark_time_utils

DBRICKS_MULTI_CLUSTER_WRITES_ENABLED = "spark.databricks.delta.multiClusterWrites.enabled"
DBRICKS_RUNTIME_VERSION = "DATABRICKS_RUNTIME_VERSION"

SPARK31_DELTA_LOGSTORE_CLASS = "spark.delta.logStore.class"
SPARK31_DYNAMODB_LOGSTORE_CLASS = "io.delta.storage.DynamoDBLogStore"

SPARK32_OR_HIGHER_DELTA_LOGSTORE_CLASS = "spark.delta.logStore.s3.impl"
SPARK32_OR_HIGHER_DYNAMODB_LOGSTORE_CLASS = "io.delta.storage.S3DynamoDBLogStore"

logger = get_logger("offline_store")


@dataclass
class OfflineStoreWriterParams:
    s3_path: str

    always_store_anchor_column: bool
    """Whether the anchor column should be stored in the Offline Feature Store regardless of whether it is
    required by the storage layer.

    If this is false the anchor column will be dropped from the stored data if it's not needed by the
    OfflineStoreWriter implementation.
    """

    time_column: str
    """The column containing the timestamp value used for time-based partitioning"""

    join_key_columns: List[str]

    is_continuous: bool


def get_offline_store_writer(
    params: OfflineStoreWriterParams, fd: FeatureDefinition, version: int, spark: SparkSession
):
    """Creates a concrete implementation of OfflineStoreWriter based on fv_config."""
    fv_config = fd.offline_store_config
    case = fv_config.WhichOneof("store_type")
    if case == "delta":
        return DeltaWriter(fd, params, spark, version)
    elif case == "parquet":
        return ParquetWriter(fd, params, spark, version)
    # Remove default after database migration is complete.
    # raise KeyError(case)
    return ParquetWriter(fd, params, spark, version)


def get_offline_store_reader(spark: SparkSession, fd: FeatureDefinition) -> "OfflineStoreReader":
    case = fd.offline_store_config.WhichOneof("store_type")
    if case == "delta":
        return DeltaReader(spark, fd)
    elif case == "parquet":
        return ParquetReader(spark, fd)
    # Remove default after database migration is complete.
    # raise KeyError(case)
    return ParquetReader(spark, fd)


class OfflineStoreWriter(ABC):
    """Interface for Offline Feature Store writers."""

    @abstractmethod
    def append_dataframe(self, data_frame):
        """Append the rows from data_frame to the Store table. Nothing is overwritten."""
        raise NotImplementedError

    @abstractmethod
    def upsert_dataframe(self, data_frame):
        """Upsert the rows from data_frame to the Store table.

        Rows with matching join keys and time column are overwritten. Other rows are inserted.
        """
        raise NotImplementedError

    @abstractmethod
    def delete_keys(self, data_frame) -> int:
        """Delete rows from the Store table that match the keys inside the data_frame.

        Return number of successfully deleted keys."""
        raise NotImplementedError


class OfflineStoreReader(ABC):
    @abstractmethod
    def read(self, time_limits: pendulum.Period) -> DataFrame:
        raise NotImplementedError


class ParquetWriter(OfflineStoreWriter):
    """Parquet implementation of OfflineStoreWriter"""

    def __init__(self, fd: FeatureDefinition, params: OfflineStoreWriterParams, spark: SparkSession, version: int):
        _check_supported_offline_store_version(fd)
        self._fd = fd
        self._params = params
        self._spark = spark
        self._version = version
        self._partition_size = partition_size_for_parquet(fd).as_timedelta()
        self._partition_col = partition_col_for_parquet(fd)

    def append_dataframe(self, data_frame):
        if self._params.is_continuous or self._fd.offline_store_config is not None:
            align_duration = core_time_utils.convert_timedelta_for_version(self._partition_size, self._version)
            aligned_time = _align_timestamp(functions.col(ANCHOR_TIME), functions.lit(align_duration))
            data_frame = data_frame.withColumn(TIME_PARTITION, aligned_time)

        data_frame.write.option("partitionOverwriteMode", "dynamic").partitionBy(self._partition_col).parquet(
            self._params.s3_path, mode="overwrite"
        )

    def upsert_dataframe(self, data_frame):
        raise NotImplementedError()

    def delete_keys(self, data_frame):
        raise NotImplementedError()


class ParquetReader(OfflineStoreReader):
    def __init__(self, spark: SparkSession, fd: FeatureDefinition):
        _check_supported_offline_store_version(fd)
        self._spark = spark
        assert fd.materialization_enabled and fd.writes_to_offline_store
        self._path = fd.materialized_data_path
        self._partition_col = partition_col_for_parquet(fd)
        self._partition_size = partition_size_for_parquet(fd).as_timedelta()
        self.version = fd.get_feature_store_format_version

    def read(self, time_limits: Optional[pendulum.Period]):
        spark_df = self._spark.read.parquet(self._path)

        # Parquet is partitioned by TIME_PARTITION when is_continuous and ANCHOR_TIME when not.
        # We want to explicitly cast the partition type in case:
        #   `spark.sql.sources.partitionColumnTypeInference.enabled` = "false"

        spark_df = spark_df.withColumn(self._partition_col, functions.col(self._partition_col).cast("long"))

        if time_limits and self._partition_size:
            aligned_start_time = core_time_utils.align_time_downwards(time_limits.start, self._partition_size)
            aligned_end_time = core_time_utils.align_time_downwards(time_limits.end, self._partition_size)
            start_time_epoch = core_time_utils.convert_timestamp_for_version(aligned_start_time, self.version)
            end_time_epoch = core_time_utils.convert_timestamp_for_version(aligned_end_time, self.version)
            partition_col = functions.col(self._partition_col)
            spark_df = spark_df.where((start_time_epoch <= partition_col) & (partition_col <= end_time_epoch))

        return spark_df.drop(TIME_PARTITION)


_EXCEPTION_PACKAGES = {
    "com.databricks.sql.transaction.tahoe",  # Used by Databricks
    "org.apache.spark.sql.delta",  # Used by open source
}

_EXCEPTION_CLASSES = {
    "ConcurrentAppendException",
    "ConcurrentDeleteReadException",
    "ConcurrentDeleteDeleteException",
    "ProtocolChangedException",  # This can occur when two txns create the same table concurrently
}

_RETRYABLE_DELTA_EXCEPTIONS = {
    f"{pkg}.{cls}" for pkg, cls in itertools.product(_EXCEPTION_PACKAGES, _EXCEPTION_CLASSES)
}


def _with_delta_retries(f, max_retries=5):
    """Retries the wrapped function upon Deltalake conflict errors."""

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        from delta.exceptions import (
            ConcurrentAppendException,
            ConcurrentDeleteDeleteException,
            ConcurrentDeleteReadException,
            ConcurrentTransactionException,
            DeltaConcurrentModificationException,
            MetadataChangedException,
            ProtocolChangedException,
        )

        final_exception = None
        for i in range(max_retries):
            try:
                if i > 0:
                    # Add a random delay (with exponential backoff) before the retries to decrease
                    # the chance of recurrent conflicts between the parallel offline store writers.
                    # Possible 10s of seconds of delay is insignificant to the overall job's latency.
                    exponential_coef = 2 ** (i - 1)
                    retry_delay = exponential_coef * random.uniform(0, 1)
                    time.sleep(retry_delay)
                f(*args, **kwargs)
                return
            except Py4JJavaError as e:
                exception_class = e.java_exception.getClass().getCanonicalName()
                if exception_class not in _RETRYABLE_DELTA_EXCEPTIONS:
                    raise e
                final_exception = e
                logger.info(
                    f"Delta transaction failed (attempt {i + 1}/5); retrying",
                    exc_info=True,  # Include information about the exception currently being handled
                )
            except (
                ConcurrentAppendException,
                ConcurrentDeleteDeleteException,
                ConcurrentDeleteReadException,
                ConcurrentTransactionException,
                DeltaConcurrentModificationException,
                MetadataChangedException,
                ProtocolChangedException,
            ) as e:
                final_exception = e
                logger.info(
                    f"Delta transaction failed (attempt {i + 1}/5); retrying",
                    exc_info=True,  # Include information about the exception currently being handled
                )
            except Exception:
                logger.warning("Uncaught exception raised during Delta write", exc_info=True)
                raise
        raise Exception(f"Exceeded maximum Delta transaction retries ({max_retries})") from final_exception

    return wrapper


def _assert_safe_delta_write_configuration(spark: SparkSession):
    """Asserts that the Spark configuration is such that it is safe to write to Delta concurrently.

    With the Open Source Delta JAR installed (as it is on EMR), writing to a Delta table concurrently with another
    Spark cluster could corrupt the table unless the Delta Logstore class is overridden.

    On Databricks everything is fine as multi-cluster writes are enabled (the default).
    """
    configs = {
        DBRICKS_RUNTIME_VERSION: os.environ.get(DBRICKS_RUNTIME_VERSION, None),
        DBRICKS_MULTI_CLUSTER_WRITES_ENABLED: spark.conf.get(DBRICKS_MULTI_CLUSTER_WRITES_ENABLED, None),
        SPARK31_DELTA_LOGSTORE_CLASS: spark.conf.get(SPARK31_DELTA_LOGSTORE_CLASS, None),
        SPARK32_OR_HIGHER_DELTA_LOGSTORE_CLASS: spark.conf.get(SPARK32_OR_HIGHER_DELTA_LOGSTORE_CLASS, None),
    }
    if configs[DBRICKS_RUNTIME_VERSION] and configs[DBRICKS_MULTI_CLUSTER_WRITES_ENABLED] == "true":
        return True

    # either the spark 3.1 or spark 3.2+ DELTA_LOGSTORE_CLASS name can be set
    if configs[SPARK31_DELTA_LOGSTORE_CLASS] == SPARK31_DYNAMODB_LOGSTORE_CLASS:
        return True
    if configs[SPARK32_OR_HIGHER_DELTA_LOGSTORE_CLASS] == SPARK32_OR_HIGHER_DYNAMODB_LOGSTORE_CLASS:
        return True
    raise AssertionError(f"Configuration is not safe for concurrent writes: {configs}")


class DeltaWriter(OfflineStoreWriter):
    """DeltaLake implementation of OfflineStoreWriter"""

    def __init__(self, fd: FeatureDefinition, params: OfflineStoreWriterParams, spark: SparkSession, version: int):
        _check_supported_offline_store_version(fd)
        self._params = params
        self._version = version
        self._spark = spark
        self._partition_size = partition_size_for_delta(fd).as_timedelta()
        self._metadata_writer = DeltaMetadataWriter(spark)
        if not spark.conf.get("spark.databricks.delta.commitInfo.userMetadata"):
            raise AssertionError(f"Expected spark.databricks.delta.commitInfo.userMetadata to be set for delta writes")

    def append_dataframe(self, data_frame: DataFrame):
        data_frame = self._add_partition(data_frame)
        self._ensure_table_exists(self._spark, data_frame.schema)
        self._append_dataframe(data_frame)

    def upsert_dataframe(self, data_frame):
        # See https://github.com/delta-io/delta/issues/282 for why this isn't at the top of the file
        from delta.tables import DeltaTable

        _assert_safe_delta_write_configuration(self._spark)

        data_frame = self._add_partition(data_frame)
        self._ensure_table_exists(self._spark, data_frame.schema)

        table = DeltaTable.forPath(self._spark, self._params.s3_path)

        base = table.toDF().alias("base")
        updates = data_frame.alias("updates")

        # Build a condition which matches on all join keys, the timestamp, and the time partition column. The time
        # partition column is not needed for correctness, but it allows some files to be skipped by Delta.
        all_match_keys = [self._params.time_column, TIME_PARTITION, *self._params.join_key_columns]
        key_matches = [base[k] == updates[k] for k in all_match_keys]
        match_condition = functools.reduce(lambda l, r: l & r, key_matches)

        @_with_delta_retries
        def _execute():
            table.merge(updates, match_condition).whenMatchedUpdateAll().whenNotMatchedInsertAll().execute()

        _execute()

    def delete_keys(self, data_frame) -> int:
        # See https://github.com/delta-io/delta/issues/282 for why this isn't at the top of the file
        from delta.tables import DeltaTable

        _assert_safe_delta_write_configuration(self._spark)

        deltaTable = DeltaTable.forPath(self._spark, self._params.s3_path)
        query = ""
        columns = data_frame.columns
        for column in columns:
            if query:
                query = query + " AND "
            query = query + "t." + column + " = k." + column

        @_with_delta_retries
        def _execute():
            deltaTable.alias("t").merge(data_frame.alias("k"), query).whenMatchedDelete().execute()

        @_with_delta_retries
        def _vacuum():
            deltaTable.vacuum()

        _execute()
        _vacuum()

        last_operation = deltaTable.history(1).collect()
        if not last_operation:
            return 0

        return int(last_operation[0].operationMetrics.get("numTargetRowsDeleted", 0))

    def delete_time_range(
        self,
        feature_start_time: datetime,
        feature_end_time: datetime,
        feature_store_format_version: int,
    ):
        from delta.tables import DeltaTable

        aligned_start_time = core_time_utils.align_time_downwards(feature_start_time, self._partition_size)
        aligned_end_time = core_time_utils.align_time_downwards(feature_end_time, self._partition_size)
        start_partition = _datetime_to_partition_str(aligned_start_time, self._partition_size)
        end_partition = _datetime_to_partition_str(aligned_end_time, self._partition_size)
        start_time = core_time_utils.convert_timestamp_for_version(feature_start_time, feature_store_format_version)
        end_time = core_time_utils.convert_timestamp_for_version(feature_end_time, feature_store_format_version)

        assert self._params.always_store_anchor_column

        table = DeltaTable.forPath(self._spark, self._params.s3_path)

        @_with_delta_retries
        def _execute():
            table.delete(
                f'{ANCHOR_TIME} >= {start_time} and {ANCHOR_TIME} < {end_time} and {TIME_PARTITION} >= to_timestamp("{start_partition}") and {TIME_PARTITION} <= to_timestamp("{end_partition}")'
            )

        _execute()

    def _add_partition(self, data_frame: DataFrame) -> DataFrame:
        """Adds the time_partition column and drops the _anchor_time column if needed."""
        partition = _timestamp_to_partition_column(
            data_frame, self._params.time_column, self._partition_size, self._version
        )
        data_frame = data_frame.withColumn(TIME_PARTITION, partition)
        if not self._params.always_store_anchor_column:
            data_frame = data_frame.drop(ANCHOR_TIME)
        return data_frame

    def _ensure_table_exists(self, spark: SparkSession, schema: StructType):
        """Ensures that the table exists with the given schema.

        Some operations (including merge) fail when the table doesn't already exist. Others (append) can have conflicts
        where they wouldn't normally when they also create a new table. This function ensures neither will happen.
        """
        df = spark.createDataFrame([], schema)  # DF with 0 rows
        self._append_dataframe(df)
        # we set auto manifest so each job generates its own manifest (necessary for athena retrieval)
        self._metadata_writer.set_table_property(
            self._params.s3_path, "delta.compatibility.symlinkFormatManifest.enabled", "true"
        )

    @_with_delta_retries
    def _append_dataframe(self, df: DataFrame):
        _assert_safe_delta_write_configuration(self._spark)
        df.write.partitionBy(TIME_PARTITION).format("delta").mode("append").save(self._params.s3_path)


class DeltaMetadataWriter:
    def __init__(self, spark: SparkSession):
        self._spark = spark

    @_with_delta_retries
    def generate_symlink_manifest(self, path: str):
        _assert_safe_delta_write_configuration(self._spark)
        # we need spark_catalog in cases where the data source switches catalogs
        self._spark.sql(f"GENERATE symlink_format_manifest FOR TABLE spark_catalog.delta.`{path}`")

    @_with_delta_retries
    def set_table_property(self, path, key, val):
        _assert_safe_delta_write_configuration(self._spark)
        # we need spark_catalog in cases where the data source switches catalogs
        existing_tbl_properties = self._spark.sql(f"show tblproperties spark_catalog.delta.`{path}`").collect()
        # tblproperties are case sensitive
        has_tbl_property = any(key == r.key and val == r.value for r in existing_tbl_properties)
        # we only set it if not already set to avoid delta conflicts
        if not has_tbl_property:
            self._spark.sql(f"ALTER TABLE spark_catalog.delta.`{path}` SET TBLPROPERTIES({key}={val})")


class DeltaReader(OfflineStoreReader):
    def __init__(self, spark: SparkSession, fd: FeatureDefinition):
        _check_supported_offline_store_version(fd)
        self._spark = spark
        assert fd.materialization_enabled and fd.writes_to_offline_store
        self._path = fd.materialized_data_path
        self._partition_size = partition_size_for_delta(fd).as_timedelta()

    def read(self, time_limits: Optional[pendulum.Period]):
        spark_df = self._spark.read.format("delta").load(self._path)

        # Whenever the partition filtering logic is changed, also make sure the changes are applied to the sql based
        # version in query/nodes.py

        # Delta is always partitioned by TIME_PARTITION. We want to explicitly cast to a timestamp in case:
        #   `spark.sql.sources.partitionColumnTypeInference.enabled` = "false"
        spark_df = spark_df.withColumn(TIME_PARTITION, functions.col(TIME_PARTITION).cast("timestamp"))

        if time_limits is not None and self._partition_size:
            aligned_start_time = core_time_utils.align_time_downwards(time_limits.start, self._partition_size)
            aligned_end_time = core_time_utils.align_time_downwards(time_limits.end, self._partition_size)
            start_partition = _datetime_to_partition_str(aligned_start_time, self._partition_size)
            end_partition = _datetime_to_partition_str(aligned_end_time, self._partition_size)
            partition_col = functions.col(TIME_PARTITION)
            spark_df = spark_df.where((start_partition <= partition_col) & (partition_col <= end_partition))

        return spark_df.drop(TIME_PARTITION)


def _timestamp_to_partition_column(df: DataFrame, time_col: str, partition_size: timedelta, version: int) -> Column:
    # For some insane reason from_unixtime returns a timestamp in the session timezone, so it's pretty annoying to
    # convert a unix time to a formatted UTC timestamp unless the session is set to UTC. This only runs in
    # materialization so we can just assert that that's the case.
    tz = df.sql_ctx.sparkSession.conf.get("spark.sql.session.timeZone")
    if tz not in {"UTC", "Etc/UTC", "GMT"}:
        raise AssertionError(f"spark.sql.session.timeZone must be UTC, not {tz}")
    time_column_type = df.schema[time_col].dataType
    allowed_types = {IntegerType(), TimestampType(), LongType()}
    if time_column_type not in allowed_types:
        raise AssertionError(f"timestamp column must be one of {allowed_types}, not {time_column_type}")
    time_val = functions.col(time_col).cast(LongType())
    if time_col == ANCHOR_TIME:
        time_val = spark_time_utils.convert_epoch_to_datetime(functions.col(time_col), version).cast(LongType())
    aligned = functions.from_unixtime(_align_timestamp(time_val, window_size_seconds(partition_size)))
    partition_format = _timestamp_formats(partition_size).spark_format
    return functions.date_format(aligned.cast(TimestampType()), partition_format)


def _datetime_to_partition_str(dt: datetime, partition_size: timedelta) -> str:
    partition_format = _timestamp_formats(partition_size).python_format
    return dt.strftime(partition_format)


def _align_timestamp(int_timestamp_col, window_size):
    return int_timestamp_col - (int_timestamp_col % window_size)
